"""
The setting fo Superb SD

Authors:
  * Jiatong Shi 2021
  * Leo 2022
"""

from dataclasses import dataclass
from pathlib import Path

from omegaconf import MISSING

from s3prl.dataio.dataset import DiarizationDataset, get_info
from s3prl.dataio.sampler import FixedBatchSizeBatchSampler, GroupSameItemSampler
from s3prl.nn.rnn import SuperbDiarizationModel

from .run import Diarization
from .util import kaldi_dir_to_csv

__all__ = [
    "SuperbSD",
]


class SuperbSD(Diarization):
    def default_config(self):
        return dict(
            start=0,
            stop=None,
            target_dir=MISSING,
            cache_dir=None,
            remove_all_cache=False,
            prepare_data=dict(
                data_dir=MISSING,
            ),
            build_dataset=dict(
                chunk_size=2000,
                subsampling=1,
                rate=16000,
                use_last_samples=True,
                label_delay=0,
            ),
            build_batch_sampler=dict(
                train=dict(
                    batch_size=8,
                    shuffle=True,
                ),
                valid=dict(
                    batch_size=1,
                ),
            ),
            build_upstream=dict(
                name=MISSING,
            ),
            build_featurizer=dict(
                layer_selections=None,
                normalize=False,
            ),
            build_downstream=dict(
                hidden_size=512,
                rnn_layers=1,
            ),
            build_model=dict(
                upstream_trainable=False,
            ),
            build_optimizer=dict(
                name="Adam",
                conf=dict(
                    lr=1.0e-4,
                ),
            ),
            build_scheduler=dict(
                name="ExponentialLR",
                gamma=0.9,
            ),
            save_model=dict(
                extra_conf=dict(
                    build_downstream_conf="${build_downstream}"
                ),  # This is redundant for ASR. Just to show how to clone other fields
            ),
            save_task=dict(),
            train=dict(
                total_steps=30000,
                log_step=500,
                eval_step=500,
                save_step=500,
                gradient_clipping=1.0,
                gradient_accumulate=4,
                valid_metric="der",
                valid_higher_better=False,
                auto_resume=True,
                resume_ckpt_dir=None,
            ),
            scoring=dict(
                thresholds=[0.3, 0.4, 0.5, 0.6, 0.7],
                median_filters=[1, 11],
            ),
        )

    def prepare_data(
        self, prepare_data: dict, target_dir: str, cache_dir: str, get_path_only=False
    ):
        """
        Prepare the task-specific data metadata (path, labels...).

        Args:
            prepare_data (dict): same in :obj:`default_config`

                ====================  ====================
                key                   description
                ====================  ====================
                data_dir              (str) - the standard Kaldi data directory
                ====================  ====================

            target_dir (str): Parse your corpus and save the csv file into this directory
            cache_dir (str): If the parsing or preprocessing takes too long time, you can save
                the temporary files into this directory. This directory is expected to be shared
                across different training sessions (different hypers and :code:`target_dir`)
            get_path_only (str): Directly return the filepaths no matter they exist or not.

        Returns:
            tuple

            1. train_path (str)
            2. valid_path (str)
            3. test_paths (List[str])

            Each path (str) should be a csv file containing the following columns:

            ====================  ====================
            column                description
            ====================  ====================
            record_id             (str) - the id for the recording
            duration              (float) - the total seconds of the recording
            wav_path              (str) - the absolute path of the recording
            utt_id                (str) - the id for the segmented utterance, should be \
                                    globally unique across all recordings instead of just \
                                    unique in a recording
            speaker               (str) - the speaker label for the segmented utterance
            start_sec             (float) - segment start second in the recording
            end_sec               (float) - segment end second in the recording
            ====================  ====================

            Instead of one waveform file per row, the above file format is one segment per row,
            and a waveform file can have multiple overlapped segments uttered by different speakers.
        """

        @dataclass
        class Config:
            data_dir: str

        conf = Config(**prepare_data)

        target_dir: Path = Path(target_dir)
        train_csv = target_dir / "train.csv"
        valid_csv = target_dir / "valid.csv"
        test_csv = target_dir / "test.csv"

        if get_path_only:
            return train_csv, valid_csv, [test_csv]

        kaldi_dir_to_csv(Path(conf.data_dir) / "train", train_csv)
        kaldi_dir_to_csv(Path(conf.data_dir) / "dev", valid_csv)
        kaldi_dir_to_csv(Path(conf.data_dir) / "test", test_csv)
        return train_csv, valid_csv, [test_csv]

    def build_dataset(
        self,
        build_dataset: dict,
        target_dir: str,
        cache_dir: str,
        mode: str,
        data_csv: str,
        data_dir: str,
        num_speakers: int,
        frame_shift: int,
    ):
        """
        Build the dataset for train/valid/test.

        Args:
            build_dataset (dict): same in :obj:`default_config`, supports arguments for :obj:`DiarizationDataset`
            target_dir (str): Current experiment directory
            cache_dir (str): If the preprocessing takes too long time, you can save
                the temporary files into this directory. This directory is expected to be shared
                across different training sessions (different hypers and :code:`target_dir`)
            mode (str): train/valid/test
            data_csv (str): The metadata csv file for the specific :code:`mode`
            data_dir (str): The converted kaldi data directory from :code:`data_csv`
            num_speakers (int): The number of speaker per utterance
            frame_shift (int): The frame shift of the upstream model (downsample rate from 16 KHz)

        Returns:
            torch Dataset

            For all train/valid/test mode, the dataset should return each item as a dictionary
            containing the following keys:

            ====================  ====================
            key                   description
            ====================  ====================
            x                     (torch.FloatTensor) - the waveform in (seq_len, 1)
            x_len                 (int) - the waveform length :code:`seq_len`
            label                 (torch.LongTensor) - the binary label for each upstream frame, \
                                    shape: :code:`(upstream_len, 2)`
            label_len             (int) - the upstream feature's seq length :code:`upstream_len`
            record_id             (str) - the unique id for the recording
            chunk_id              (int) - since recording can be chunked into several segments \
                                    for efficient training, this field indicate the segment's \
                                    original position (order, 0-index) in the recording. This \
                                    field is only useful during the testing stage
            ====================  ====================
        """

        dataset = DiarizationDataset(
            mode,
            data_dir,
            frame_shift=frame_shift,
            num_speakers=num_speakers,
            **build_dataset,
        )
        return dataset

    def build_batch_sampler(
        self,
        build_batch_sampler: dict,
        target_dir: str,
        cache_dir: str,
        mode: str,
        data_csv: str,
        data_dir: str,
        dataset,
    ):
        """
        Return the batch sampler for torch DataLoader.

        Args:
            build_batch_sampler (dict): same in :obj:`default_config`

                ====================  ====================
                key                   description
                ====================  ====================
                train                 (dict) - arguments for :obj:`FixedBatchSizeBatchSampler`
                valid                 (dict) - arguments for :obj:`FixedBatchSizeBatchSampler`
                test                  (dict) - arguments for :obj:`GroupSameItemSampler`, should always \
                                        use this batch sampler for the testing stage
                ====================  ====================

            target_dir (str): Current experiment directory
            cache_dir (str): If the preprocessing takes too long time, save
                the temporary files into this directory. This directory is expected to be shared
                across different training sessions (different hypers and :code:`target_dir`)
            mode (str): train/valid/test
            data_csv (str): The metadata csv file for the specific :code:`mode`
            data_dir (str): The converted kaldi data directory from :code:`data_csv`
            dataset: the dataset from :obj:`build_dataset`

        Returns:
            batch sampler for torch DataLoader
        """

        @dataclass
        class Config:
            train: dict = None
            valid: dict = None

        conf = Config(**build_batch_sampler)

        if mode == "train":
            return FixedBatchSizeBatchSampler(dataset, **(conf.train or {}))
        elif mode == "valid":
            return FixedBatchSizeBatchSampler(dataset, **(conf.valid or {}))
        elif mode == "test":
            record_ids = get_info(dataset, ["record_id"])
            return GroupSameItemSampler(record_ids)
        else:
            raise ValueError(f"Unsupported mode: {mode}")

    def build_downstream(
        self,
        build_downstream: dict,
        downstream_input_size: int,
        downstream_output_size: int,
        downstream_input_stride: int,
    ):
        """
        Return the task-specific downstream model.
        By default build the :obj:`SuperbDiarizationModel` model

        Args:
            build_downstream (dict): same in :obj:`default_config`, support arguments of :obj:`SuperbDiarizationModel`
            downstream_input_size (int): the required input size of the model
            downstream_output_size (int): the required output size of the model
            downstream_input_stride (int): the input feature's stride (from 16 KHz)

        Returns:
            :obj:`s3prl.nn.interface.AbsFrameModel`
        """
        return SuperbDiarizationModel(
            downstream_input_size, downstream_output_size, **build_downstream
        )



# --------------------------------------------------------
# LightHuBERT: Lightweight and Configurable Speech Representation Learning with Once-for-All Hidden-Unit BERT (https://arxiv.org/pdf/2203.15610.pdf)
# Github source: https://github.com/mechanicalsea/lighthubert
# Licensed under The MIT License [see LICENSE for details]
# Based on fairseq code bases
# https://github.com/pytorch/fairseq
# --------------------------------------------------------

import torch
import torch.nn as nn
import torch.nn.functional as F


class SamePad(nn.Module):
    def __init__(self, kernel_size, causal=False):
        super().__init__()
        if causal:
            self.remove = kernel_size - 1
        else:
            self.remove = 1 if kernel_size % 2 == 0 else 0

    def forward(self, x):
        if self.remove > 0:
            x = x[:, :, : -self.remove]
        return x


class TransposeLast(nn.Module):
    def __init__(self, deconstruct_idx=None):
        super().__init__()
        self.deconstruct_idx = deconstruct_idx

    def forward(self, x):
        if self.deconstruct_idx is not None:
            x = x[self.deconstruct_idx]
        return x.transpose(-2, -1)


class Fp32LayerNorm(nn.LayerNorm):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def forward(self, input):
        output = F.layer_norm(
            input.float(),
            self.normalized_shape,
            self.weight.float() if self.weight is not None else None,
            self.bias.float() if self.bias is not None else None,
            self.eps,
        )
        return output.type_as(input)


class Fp32GroupNorm(nn.GroupNorm):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def forward(self, input):
        output = F.group_norm(
            input.float(),
            self.num_groups,
            self.weight.float() if self.weight is not None else None,
            self.bias.float() if self.bias is not None else None,
            self.eps,
        )
        return output.type_as(input)


def quant_noise(module, p, block_size):
    """
    Wraps modules and applies quantization noise to the weights for
    subsequent quantization with Iterative Product Quantization as
    described in "Training with Quantization Noise for Extreme Model Compression"

    Args:
        - module: nn.Module
        - p: amount of Quantization Noise
        - block_size: size of the blocks for subsequent quantization with iPQ

    Remarks:
        - Module weights must have the right sizes wrt the block size
        - Only Linear, Embedding and Conv2d modules are supported for the moment
        - For more detail on how to quantize by blocks with convolutional weights,
          see "And the Bit Goes Down: Revisiting the Quantization of Neural Networks"
        - We implement the simplest form of noise here as stated in the paper
          which consists in randomly dropping blocks
    """

    # if no quantization noise, don't register hook
    if p <= 0:
        return module

    # supported modules
    assert isinstance(module, (nn.Linear, nn.Embedding, nn.Conv2d))

    # test whether module.weight has the right sizes wrt block_size
    is_conv = module.weight.ndim == 4

    # 2D matrix
    if not is_conv:
        assert (
            module.weight.size(1) % block_size == 0
        ), "Input features must be a multiple of block sizes"

    # 4D matrix
    else:
        # 1x1 convolutions
        if module.kernel_size == (1, 1):
            assert (
                module.in_channels % block_size == 0
            ), "Input channels must be a multiple of block sizes"
        # regular convolutions
        else:
            k = module.kernel_size[0] * module.kernel_size[1]
            assert k % block_size == 0, "Kernel size must be a multiple of block size"

    def _forward_pre_hook(mod, input):
        # no noise for evaluation
        if mod.training:
            if not is_conv:
                # gather weight and sizes
                weight = mod.weight
                in_features = weight.size(1)
                out_features = weight.size(0)

                # split weight matrix into blocks and randomly drop selected blocks
                mask = torch.zeros(
                    in_features // block_size * out_features, device=weight.device
                )
                mask.bernoulli_(p)
                mask = mask.repeat_interleave(block_size, -1).view(-1, in_features)

            else:
                # gather weight and sizes
                weight = mod.weight
                in_channels = mod.in_channels
                out_channels = mod.out_channels

                # split weight matrix into blocks and randomly drop selected blocks
                if mod.kernel_size == (1, 1):
                    mask = torch.zeros(
                        int(in_channels // block_size * out_channels),
                        device=weight.device,
                    )
                    mask.bernoulli_(p)
                    mask = mask.repeat_interleave(block_size, -1).view(-1, in_channels)
                else:
                    mask = torch.zeros(
                        weight.size(0), weight.size(1), device=weight.device
                    )
                    mask.bernoulli_(p)
                    mask = (
                        mask.unsqueeze(2)
                        .unsqueeze(3)
                        .repeat(1, 1, mod.kernel_size[0], mod.kernel_size[1])
                    )

            # scale weights and apply mask
            mask = mask.to(
                torch.bool
            )  # x.bool() is not currently supported in TorchScript
            s = 1 / (1 - p)
            mod.weight.data = s * weight.masked_fill(mask, 0)

    module.register_forward_pre_hook(_forward_pre_hook)
    return module


class GradMultiply(torch.autograd.Function):
    @staticmethod
    def forward(ctx, x, scale):
        ctx.scale = scale
        res = x.new(x)
        return res

    @staticmethod
    def backward(ctx, grad):
        return grad * ctx.scale, None



# --------------------------------------------------------
# LightHuBERT: Lightweight and Configurable Speech Representation Learning with Once-for-All Hidden-Unit BERT (https://arxiv.org/pdf/2203.15610.pdf)
# Github source: https://github.com/mechanicalsea/lighthubert
# Licensed under The MIT License [see LICENSE for details]
# Based on fairseq code bases
# https://github.com/pytorch/fairseq
# --------------------------------------------------------

import math
import warnings
from typing import Callable

import torch
import torch.nn.functional as F


def is_xla_tensor(tensor):
    return torch.is_tensor(tensor) and tensor.device.type == "xla"


def index_put(tensor, indices, value):
    if is_xla_tensor(tensor):
        for _ in range(indices.dim(), tensor.dim()):
            indices = indices.unsqueeze(-1)
        if indices.size(-1) < tensor.size(-1):
            indices = indices.expand_as(tensor)
        tensor = torch.mul(tensor, ~indices) + torch.mul(value, indices)
    else:
        tensor[indices] = value
    return tensor


def pad_to_multiple(x, multiple, dim=-1, value=0):
    # Inspired from https://github.com/lucidrains/local-attention/blob/master/local_attention/local_attention.py#L41
    if x is None:
        return None, 0
    tsz = x.size(dim)
    m = tsz / multiple
    remainder = math.ceil(m) * multiple - tsz
    if m.is_integer():
        return x, 0
    pad_offset = (0,) * (-1 - dim) * 2

    return F.pad(x, (*pad_offset, 0, remainder), value=value), remainder


def gelu_accurate(x):
    if not hasattr(gelu_accurate, "_a"):
        gelu_accurate._a = math.sqrt(2 / math.pi)
    return (
        0.5 * x * (1 + torch.tanh(gelu_accurate._a * (x + 0.044715 * torch.pow(x, 3))))
    )


def gelu(x: torch.Tensor) -> torch.Tensor:
    return torch.nn.functional.gelu(x.float()).type_as(x)


def relu_squared(x: torch.Tensor):
    return F.relu(x).pow(2)


def deprecation_warning(message, stacklevel=3):
    # don't use DeprecationWarning, since it's ignored by default
    warnings.warn(message, stacklevel=stacklevel)


def get_activation_fn(activation: str) -> Callable:
    """Returns the activation function corresponding to `activation`"""

    def gelu_accurate(x):
        if not hasattr(gelu_accurate, "_a"):
            gelu_accurate._a = math.sqrt(2 / math.pi)
        return (
            0.5
            * x
            * (1 + torch.tanh(gelu_accurate._a * (x + 0.044715 * torch.pow(x, 3))))
        )

    def gelu(x: torch.Tensor) -> torch.Tensor:
        return torch.nn.functional.gelu(x.float()).type_as(x)

    if activation == "relu":
        return F.relu
    elif activation == "relu_squared":
        return relu_squared
    elif activation == "gelu":
        return gelu
    elif activation == "gelu_fast":
        deprecation_warning(
            "--activation-fn=gelu_fast has been renamed to gelu_accurate"
        )
        return gelu_accurate
    elif activation == "gelu_accurate":
        return gelu_accurate
    elif activation == "tanh":
        return torch.tanh
    elif activation == "linear":
        return lambda x: x
    elif activation == "swish":
        return torch.nn.SiLU
    else:
        raise RuntimeError("--activation-fn {} not supported".format(activation))



"""
Control how torch DataLoader group instances into a batch
"""

from .balanced_weighted_sampler import BalancedWeightedSampler
from .distributed_sampler import DistributedBatchSamplerWrapper
from .fixed_batch_size_batch_sampler import FixedBatchSizeBatchSampler
from .group_same_item_sampler import GroupSameItemSampler
from .max_timestamp_batch_sampler import MaxTimestampBatchSampler
from .sorted_sampler import SortedBucketingSampler, SortedSliceSampler

__all__ = [
    "BalancedWeightedSampler",
    "DistributedBatchSamplerWrapper",
    "FixedBatchSizeBatchSampler",
    "GroupSameItemSampler",
    "MaxTimestampBatchSampler",
    "SortedBucketingSampler",
    "SortedSliceSampler",
]



import copy
from dataclasses import dataclass

import torch

from .base import AugmentedDynamicItemDataset, DataPipe


@dataclass
class AutoregressivePrediction(DataPipe):
    n_future: int = 5
    source_feat_name: str = (
        "source_feat"  # tensors in the shape of: (seq_len, feat_dim)
    )
    target_feat_name: str = (
        "target_feat"  # tensors in the shape of: (seq_len, feat_dim)
    )
    source_feat_len_name: str = "feat_len"

    def generate_shifted_data(self, source_feat):
        with torch.no_grad():
            feat_len = int(source_feat.size(0)) - self.n_future
            target_feat = copy.deepcopy(source_feat[self.n_future :, :])
            source_feat = source_feat[: -self.n_future, :]

            target_feat = target_feat.to(dtype=torch.float32)
            source_feat = source_feat.to(dtype=torch.float32)

        return source_feat, target_feat, feat_len

    def __call__(self, dataset: AugmentedDynamicItemDataset):
        dataset.add_dynamic_item(
            self.generate_shifted_data,
            takes=self.source_feat_name,
            provides=[
                self.source_feat_name,
                self.target_feat_name,
                self.source_feat_len_name,
            ],
        )
        return dataset



from collections import Counter

import pytest

from s3prl.dataio.sampler import BalancedWeightedSampler


@pytest.mark.parametrize("duplicate", [10000, 100000])
def test_balanced_weighted_sampler(duplicate: int):
    labels = ["a", "a", "b", "a"]
    batch_size = 5
    prev_diff_ratio = 1.0
    sampler = BalancedWeightedSampler(
        labels, batch_size=batch_size, duplicate=duplicate, seed=0
    )
    indices = list(sampler)
    assert len(indices[0]) == batch_size

    counter = Counter()
    for batch_indices in indices:
        for idx in batch_indices:
            counter.update(labels[idx])

    diff_ratio = abs(counter["a"] - counter["b"]) / duplicate * len(labels)
    assert diff_ratio < prev_diff_ratio
    prev_diff_ratio = diff_ratio

    diff_ratio < 0.05



# -*- coding: utf-8 -*- #
"""*********************************************************************************************"""
#   FileName     [ upstream/vggish/expert.py ]
#   Synopsis     [ the VGGish wrapper ]
#   Author       [ S3PRL ]
#   Copyright    [ Copyleft(c), Speech Lab, NTU, Taiwan ]
"""*********************************************************************************************"""

from torch.nn.utils.rnn import pad_sequence

from ..interfaces import UpstreamBase
from .audio import waveform_to_examples
from .vggish import VGGish


class UpstreamExpert(UpstreamBase):
    def __init__(self, ckpt, **kwargs):
        super().__init__(**kwargs)
        self.model = VGGish(ckpt, **kwargs)

    def get_downsample_rates(self, key: str) -> int:
        return 16000

    def forward(self, wavs):
        device = wavs[0].device

        outputs = []
        for wav in wavs:
            # each example is in - (num_examples, 1, num_frames, num_bands)
            feature = waveform_to_examples(wav.detach().cpu().numpy())
            feature = self.model(feature.to(device))
            if feature.dim() == 1:
                feature = feature.unsqueeze(0)
            outputs.append(feature)
        outputs = pad_sequence(outputs, batch_first=True)

        return {
            "last_hidden_state": outputs,
            "hidden_states": [outputs],
        }



"""
Common linear models

Authors:
  * Leo 2022
"""

from .common import FrameLevel, UtteranceLevel

__all__ = [
    "FrameLevelLinear",
    "MeanPoolingLinear",
]


class FrameLevelLinear(FrameLevel):
    """
    The frame-level linear probing model used in SUPERB Benchmark
    """

    def __init__(
        self,
        input_size: int,
        output_size: int,
        hidden_size: int = 256,
    ):
        super().__init__(input_size, output_size, hidden_sizes=[hidden_size])


class MeanPoolingLinear(UtteranceLevel):
    """
    The utterance-level linear probing model used in SUPERB Benchmark
    """

    def __init__(
        self,
        input_size: int,
        output_size: int,
        hidden_size: int = 256,
    ):
        super().__init__(input_size, output_size, hidden_sizes=[hidden_size])



"""
Fix random seeds

Authors
  * Heng-Jui Chang 2022
"""

import random

import numpy as np
import torch

__all__ = [
    "fix_random_seeds",
]


def fix_random_seeds(seed: int = 1337) -> None:
    """Fixes all random seeds, including cuDNN backends.

    Args:
        seed (int, optional): Random seed. Defaults to 1337.
    """

    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    if torch.cuda.is_available():
        torch.cuda.manual_seed(seed)
        torch.cuda.manual_seed_all(seed)
    torch.backends.cudnn.benchmark = False
    torch.backends.cudnn.deterministic = True



import torch

from s3prl.nn import FrameLevel


def test_FrameLevel(helpers):
    module = FrameLevel(3, 4, [5, 6])
    x = torch.randn(32, 10, 3)
    x_len = (torch.ones(32) * 3).long()
    h, hl = module(x, x_len)



from .base import SequentialDataPipe
from .common_pipes import LoadAudio, SetOutputKeys
from .extract_feat_pipes import ExtractKaldiFeat
from .masked_reconstruction_pipes import MaskedReconstruction, PrepareTargetFeat


class PretrainMockingjayPipe(SequentialDataPipe):
    """
    each item in the input dataset should have:
        wav_path: str
    """

    def __init__(
        self,
        output_keys: dict = None,
        position_encoding_size: int = 768,
        mask_proportion: float = 0.15,
        mask_consecutive_min: int = 7,
        mask_consecutive_max: int = 7,
        mask_allow_overlap: bool = True,
        mask_bucket_ratio: float = 1.5,
        mask_frequency: int = 0.2,
        kaldi: dict = {
            "feat_type": "fbank",
            "fbank": {
                "frame_length": 25.0,
                "frame_shift": 10.0,
                "num_mel_bins": 80,  # because delta={"order": 2}
                "use_log_fbank": True,
            },
            "mfcc": {"frame_length": 25.0, "frame_shift": 10.0, "num_ceps": 13},
            "spectrogram": {"frame_length": 25.0, "frame_shift": 10.0},
        },
        delta: dict = {"order": 2, "win_length": 5},
        cmvn: dict = {"use_cmvn": True},
        audio_sample_rate: int = 16000,
        audio_channel_reduction: str = "first",
        n_jobs: int = 6,
    ):
        """
        Args:
            output_keys (dict): args for the output handle
            position_encoding_size (int): this should be identical to `hidden_size`
            mask_proportion (float): mask this percentage of all spectrogram frames in each sequence at random during MAM training
            mask_consecutive_min (int): mask this amount of consecutive frames
            mask_consecutive_max (int): mask this amount of consecutive frames
            mask_allow_overlap (bool): allow overlap masking
            mask_bucket_ratio (float): only used when overlap is not allowed. sample a mask from each bucket in size of [sampled mask_consecutive * mask_bucket_ratio]
            mask_frequency (float): mask maximum this percentage of frequency bands, set to 0 for no frequency mask
            kaldi (dict): args for the kaldi extracter
            delta (dict): args for applying delta on features
            cmvn (dict): args for applying cmvn on features
            n_mels (int): number of mel features
            n_mfcc (int): number of mfcc features
            audio_sample_rate (int): audio sample rate
            audio_channel_reduction (str): "first" channel
            n_jobs (int): number of workers
        """
        output_keys = output_keys or dict(
            x="masked_feat",
            label="target_feat",
            label_mask="label_mask",
            position_encoding="pos_enc",
            attention_mask="attn_mask",
            unique_name="id",
        )

        super().__init__(
            LoadAudio(
                n_jobs=n_jobs,
                audio_sample_rate=audio_sample_rate,
                audio_channel_reduction=audio_channel_reduction,
            ),
            ExtractKaldiFeat(
                kaldi=kaldi, delta=delta, cmvn=cmvn, feat_name="source_feat"
            ),
            PrepareTargetFeat(
                use_copy=True,
                source_feat_name="source_feat",
                target_feat_name="target_feat",
            ),
            MaskedReconstruction(
                position_encoding_size=position_encoding_size,
                mask_proportion=mask_proportion,
                mask_consecutive_min=mask_consecutive_min,
                mask_consecutive_max=mask_consecutive_max,
                mask_allow_overlap=mask_allow_overlap,
                mask_bucket_ratio=mask_bucket_ratio,
                mask_frequency=mask_frequency,
                source_feat_name="source_feat",
                target_feat_name="target_feat",
                masked_feat_name="masked_feat",
                pos_enc_name="pos_enc",
                attn_mask_name="attn_mask",
                label_mask_name="label_mask",
            ),
            SetOutputKeys(output_keys=output_keys),
        )



# -*- coding: utf-8 -*- #
"""*********************************************************************************************"""
#   FileName     [ upstream/npc/hubconf.py ]
#   Synopsis     [ the npc torch hubconf ]
#   Author       [ S3PRL ]
#   Copyright    [ Copyleft(c), Speech Lab, NTU, Taiwan ]
"""*********************************************************************************************"""


import os

from s3prl.util.download import _urls_to_filepaths

from .expert import UpstreamExpert as _UpstreamExpert


def npc_local(ckpt, *args, **kwargs):
    """
    The model from local ckpt
        ckpt (str): PATH
    """
    assert os.path.isfile(ckpt)
    return _UpstreamExpert(ckpt, *args, **kwargs)


def npc_url(ckpt, refresh=False, *args, **kwargs):
    """
    The model from URL
        ckpt (str): URL
    """
    return npc_local(_urls_to_filepaths(ckpt, refresh=refresh), *args, **kwargs)


def npc(refresh=False, *args, **kwargs):
    """
    The default model
        refresh (bool): whether to download ckpt/config again if existed
    """
    return npc_360hr(refresh=refresh, *args, **kwargs)


def npc_360hr(refresh=False, *args, **kwargs):
    """
    The npc standard model on 360hr
        refresh (bool): whether to download ckpt/config again if existed
    """
    kwargs[
        "ckpt"
    ] = "https://huggingface.co/leo19941227/apc_series/resolve/main/npc_360hr.ckpt"
    return npc_url(refresh=refresh, *args, **kwargs)


def npc_960hr(refresh=False, *args, **kwargs):
    """
    The npc standard model on 960hr
        refresh (bool): whether to download ckpt/config again if existed
    """
    kwargs[
        "ckpt"
    ] = "https://huggingface.co/leo19941227/apc_series/resolve/main/npc_960hr.ckpt"
    return npc_url(refresh=refresh, *args, **kwargs)



from .autoregressive_prediction_pipes import AutoregressivePrediction
from .base import SequentialDataPipe
from .common_pipes import LoadAudio, SetOutputKeys
from .extract_feat_pipes import ExtractApcFeat


class PretrainApcPipe(SequentialDataPipe):
    """
    each item in the input dataset should have:
        wav_path: str
    """

    def __init__(
        self,
        output_keys: dict = None,
        n_future: int = 5,
        feat_type: str = "fbank",
        feat_dim: int = 80,
        frame_length: int = 25,
        frame_shift: int = 10,
        decode_wav: bool = False,
        cmvn: bool = True,
        audio_sample_rate: int = 16000,
        audio_channel_reduction: str = "first",
        n_jobs: int = 6,
    ):
        """
        Args:
            output_keys (dict): args for the output handle
            n_future (int): number of future steps for the autoregressive predictive task
            feat_type (str): feature type
            feat_dim (int): feature dimension
            frame_length (int): window size in ms
            frame_shift (int): hop size in ms
            decode_wav (bool): whether to decode wav
            cmvn (bool): whether to apply uttr.-wised CMVN on feature
            audio_sample_rate (int): audio sample rate
            audio_channel_reduction (str): "first" channel
            n_jobs (int): number of workers
        """
        output_keys = output_keys or dict(
            x="source_feat",
            label="target_feat",
            x_len="feat_len",
            unique_name="id",
        )

        super().__init__(
            LoadAudio(
                n_jobs=n_jobs,
                audio_sample_rate=audio_sample_rate,
                audio_channel_reduction=audio_channel_reduction,
            ),
            ExtractApcFeat(
                feat_type=feat_type,
                feat_dim=feat_dim,
                frame_length=frame_length,
                frame_shift=frame_shift,
                decode_wav=decode_wav,
                cmvn=cmvn,
                feat_name="source_feat",
            ),
            AutoregressivePrediction(
                n_future=n_future,
                source_feat_name="source_feat",
                target_feat_name="target_feat",
                source_feat_len_name="feat_len",
            ),
            SetOutputKeys(output_keys=output_keys),
        )



"""
Parse VoxCeleb1 corpus for classification

Authors:
  * Leo 2022
  * Cheng Liang 2022
"""

import logging
import os
from collections import defaultdict
from pathlib import Path
from typing import List

from filelock import FileLock
from joblib import Parallel, delayed
from tqdm import tqdm

from .base import Corpus

logger = logging.getLogger(__name__)

SPLIT_FILE_URL = "https://www.robots.ox.ac.uk/~vgg/data/voxceleb/meta/iden_split.txt"
CACHE_ROOT = Path.home() / ".cache" / "s3prl"

__all__ = [
    "VoxCeleb1SID",
]


class VoxCeleb1SID(Corpus):
    def __init__(
        self, dataset_root: str, n_jobs: int = 4, cache_root: str = CACHE_ROOT
    ) -> None:
        self.dataset_root = Path(dataset_root).resolve()

        uid2split = self._get_standard_usage(self.dataset_root, cache_root)
        self._split2uids = defaultdict(list)
        for uid, split in uid2split.items():
            self._split2uids[split].append(Path(uid.replace("/", "-")).stem)

        uid2wavpath = self._find_wavs_with_uids(
            self.dataset_root, sorted(uid2split.keys()), n_jobs=n_jobs
        )
        self._data = {
            Path(uid.replace("/", "-")).stem: {
                "wav_path": uid2wavpath[uid],
                "label": self._build_label(uid),
            }
            for uid in uid2split.keys()
        }

    @property
    def all_data(self):
        return self._data

    @property
    def data_split_ids(self):
        return (
            self._split2uids["train"],
            self._split2uids["valid"],
            self._split2uids["test"],
        )

    @staticmethod
    def _get_standard_usage(dataset_root: Path, cache_root: Path):
        split_filename = SPLIT_FILE_URL.split("/")[-1]
        split_filepath = Path(cache_root) / split_filename
        if not split_filepath.is_file():
            with FileLock(str(split_filepath) + ".lock"):
                os.system(f"wget {SPLIT_FILE_URL} -O {str(split_filepath)}")
        standard_usage = [
            line.strip().split(" ") for line in open(split_filepath, "r").readlines()
        ]

        def code2split(code: int):
            splits = ["train", "valid", "test"]
            return splits[code - 1]

        standard_usage = {uid: code2split(int(split)) for split, uid in standard_usage}
        return standard_usage

    @staticmethod
    def _find_wavs_with_uids(dataset_root, uids, n_jobs=4):
        def find_wav_with_uid(uid):
            found_wavs = list(dataset_root.glob(f"*/wav/{uid}"))
            assert len(found_wavs) == 1
            return uid, found_wavs[0]

        uids_with_wavs = Parallel(n_jobs=n_jobs)(
            delayed(find_wav_with_uid)(uid) for uid in tqdm(uids, desc="Search wavs")
        )
        uids2wav = {uid: wav for uid, wav in uids_with_wavs}
        return uids2wav

    @staticmethod
    def _build_label(uid):
        id_string = uid.split("/")[0]
        label = f"speaker_{int(id_string[2:]) - 10001}"
        return label

    @classmethod
    def download_dataset(
        cls, target_dir: str, splits: List[str] = ["dev", "test"]
    ) -> None:
        tgt_dir = os.path.abspath(target_dir)
        assert os.path.exists(tgt_dir), "Target directory does not exist"

        from zipfile import ZipFile

        import requests

        def unzip_then_delete(filepath: str, split: str):
            assert os.path.exists(filepath), "File not found!"

            with ZipFile(filepath) as zipf:
                zipf.extractall(path=os.path.join(tgt_dir, "Voxceleb1", split))
            os.remove(os.path.abspath(filepath))

        def download_from_url(url: str, split: str):
            filename = url.split("/")[-1].replace(" ", "_")
            filepath = os.path.join(tgt_dir, filename)

            r = requests.get(url, stream=True)
            if r.ok:
                logger.info(f"Saving {filename} to", filepath)
                with open(filepath, "wb") as f:
                    for chunk in r.iter_content(chunk_size=1024 * 1024 * 10):
                        if chunk:
                            f.write(chunk)
                            f.flush()
                            os.fsync(f.fileno())
                logger.info(f"{filename} successfully downloaded")
            else:
                logger.info(f"Download failed: status code {r.status_code}\n{r.text}")

            return filepath

        def download_dev():
            partpaths = []

            for part in ["a", "b", "c", "d"]:
                if os.path.exists(os.path.join(tgt_dir, f"vox1_dev_wav_parta{part}")):
                    logger.info(f"vox1_dev_wav_parta{part} exists, skip donwload")
                    partpaths.append(os.path.join(tgt_dir, f"vox1_dev_wav_parta{part}"))
                    continue
                fp = download_from_url(
                    f"https://thor.robots.ox.ac.uk/~vgg/data/voxceleb/vox1a/vox1_dev_wav_parta{part}",
                    "dev",
                )
                partpaths.append(fp)

            zippath = os.path.join(tgt_dir, "vox1_dev_wav.zip")
            with open(zippath, "wb") as outfile:
                for f in partpaths:
                    with open(f, "rb") as infile:
                        for line in infile:
                            outfile.write(line)

            for f in partpaths:
                os.remove(f)
            unzip_then_delete(zippath, "dev")

        for split in splits:
            if not os.path.exists(os.path.join(tgt_dir, "Voxceleb1/" + split + "/wav")):
                if split == "dev":
                    download_dev()
                else:
                    filepath = download_from_url(
                        "https://thor.robots.ox.ac.uk/~vgg/data/voxceleb/vox1a/vox1_test_wav.zip",
                        "test",
                    )
                    unzip_then_delete(filepath, "test")

        logger.info(f"Voxceleb1 dataset downloaded. Located at {tgt_dir}/Voxceleb1/")



"""
The setting of Superb ER

Authors
  * Tzu-Hsien Huang 2021
  * Leo 2021
  * Leo 2022
"""

import json
import logging
from pathlib import Path
from typing import List

import pandas as pd
import torch
from omegaconf import MISSING
from torch.utils.data import random_split

from s3prl.dataio.corpus.iemocap import IEMOCAP
from s3prl.util.download import download

from .superb_sid import SuperbSID

logger = logging.getLogger(__name__)

__all__ = [
    "iemocap_for_superb",
    "SuperbER",
]


def iemocap_for_superb(
    target_dir: str,
    cache_dir: str,
    iemocap: str,
    test_fold: int,
    valid_ratio: float = 0.2,
    get_path_only: bool = False,
):
    """
    Prepare IEMOCAP for emotion classfication with SUPERB protocol,
    following :obj:`SuperbER.prepare_data` format.

    .. note::

        In SUPERB protocol, you need to do 5-fold cross validation.

        Also, only use 4 emotion classes: :code:`happy`, :code:`angry`,
        :code:`neutral`, and :code:`sad` with balanced data points and
        the :code:`excited` class is merged into :code:`happy` class.

    Args:
        iemocap (str): The root path of the IEMOCAP
        test_fold (int): Which fold to use as the test fold, select from 0 to 4
        valid_ratio (float): given the remaining 4 folds, how many data to use as the validation set
        **others: refer to :obj:`SuperbER.prepare_data`
    """
    target_dir = Path(target_dir)

    train_path = target_dir / f"train.csv"
    valid_path = target_dir / f"valid.csv"
    test_paths = [target_dir / f"test.csv"]

    if get_path_only:
        return train_path, valid_path, test_paths

    corpus = IEMOCAP(iemocap)
    all_datapoints = corpus.all_data

    def format_fields(data: dict):
        result = dict()
        for data_id in data.keys():
            datapoint = data[data_id]
            result[data_id] = dict(
                wav_path=datapoint["wav_path"],
                label=datapoint["emotion"],
            )
        return result

    def filter_data(data_ids: List[str]):
        result = dict()
        for data_id in data_ids:
            data_point = all_datapoints[data_id]
            if data_point["emotion"] in ["neu", "hap", "ang", "sad", "exc"]:
                if data_point["emotion"] == "exc":
                    data_point["emotion"] = "hap"
                result[data_id] = data_point
        return result

    test_session_id = test_fold + 1
    train_meta_data_json = (
        Path(cache_dir) / f"test_session{test_session_id}_train_metadata.json"
    )
    test_meta_data_json = (
        Path(cache_dir) / f"test_session{test_session_id}_test_metadata.json"
    )
    download(
        train_meta_data_json,
        f"https://huggingface.co/datasets/s3prl/iemocap_split/raw/4097f2b496c41eed016d4e5eb0ada4cccd46d1f3/Session{test_session_id}/train_meta_data.json",
        refresh=False,
    )
    download(
        test_meta_data_json,
        f"https://huggingface.co/datasets/s3prl/iemocap_split/raw/4097f2b496c41eed016d4e5eb0ada4cccd46d1f3/Session{test_session_id}/test_meta_data.json",
        refresh=False,
    )

    with open(train_meta_data_json) as f:
        metadata = json.load(f)["meta_data"]

    dev_ids = [Path(item["path"]).stem for item in metadata]

    with open(test_meta_data_json) as f:
        metadata = json.load(f)["meta_data"]

    test_ids = [Path(item["path"]).stem for item in metadata]

    train_len = int((1 - valid_ratio) * len(dev_ids))
    train_valid_lens = [train_len, len(dev_ids) - train_len]

    torch.manual_seed(0)
    train_ids, valid_ids = random_split(dev_ids, train_valid_lens)

    train_data = format_fields(filter_data(train_ids))
    valid_data = format_fields(filter_data(valid_ids))
    test_data = format_fields(filter_data(test_ids))

    def dict_to_csv(data_dict, csv_path):
        keys = sorted(list(data_dict.keys()))
        fields = sorted(data_dict[keys[0]].keys())
        data = dict()
        for field in fields:
            data[field] = []
            for key in keys:
                data[field].append(data_dict[key][field])
        data["id"] = keys
        df = pd.DataFrame(data)
        df.to_csv(csv_path, index=False)

    dict_to_csv(train_data, train_path)
    dict_to_csv(valid_data, valid_path)
    dict_to_csv(test_data, test_paths[0])

    return train_path, valid_path, test_paths


class SuperbER(SuperbSID):
    def default_config(self) -> dict:
        return dict(
            start=0,
            stop=None,
            target_dir=MISSING,
            cache_dir=None,
            remove_all_cache=False,
            prepare_data=dict(
                iemocap=MISSING,
                test_fold=MISSING,
            ),
            build_encoder=dict(),
            build_dataset=dict(),
            build_batch_sampler=dict(
                train=dict(
                    batch_size=4,
                    shuffle=True,
                ),
                valid=dict(
                    batch_size=4,
                ),
                test=dict(
                    batch_size=4,
                ),
            ),
            build_upstream=dict(
                name=MISSING,
            ),
            build_featurizer=dict(
                layer_selections=None,
                normalize=False,
            ),
            build_downstream=dict(
                hidden_size=256,
            ),
            build_model=dict(
                upstream_trainable=False,
            ),
            build_task=dict(),
            build_optimizer=dict(
                name="Adam",
                conf=dict(
                    lr=1.0e-4,
                ),
            ),
            build_scheduler=dict(
                name="ExponentialLR",
                gamma=0.9,
            ),
            save_model=dict(),
            save_task=dict(),
            train=dict(
                total_steps=30000,
                log_step=500,
                eval_step=1000,
                save_step=1000,
                gradient_clipping=1.0,
                gradient_accumulate=8,
                valid_metric="accuracy",
                valid_higher_better=True,
                auto_resume=True,
                resume_ckpt_dir=None,
            ),
            evaluate=dict(),
        )

    def prepare_data(
        self,
        prepare_data: dict,
        target_dir: str,
        cache_dir: str,
        get_path_only: bool = False,
    ):
        """
        Prepare the task-specific data metadata (path, labels...).
        By default call :obj:`iemocap_for_superb` with :code:`**prepare_data`

        Args:
            prepare_data (dict): same in :obj:`default_config`,
                support arguments in :obj:`iemocap_for_superb`
            target_dir (str): Parse your corpus and save the csv file into this directory
            cache_dir (str): If the parsing or preprocessing takes too long time, you can save
                the temporary files into this directory. This directory is expected to be shared
                across different training sessions (different hypers and :code:`target_dir`)
            get_path_only (str): Directly return the filepaths no matter they exist or not.

        Returns:
            tuple

            1. train_path (str)
            2. valid_path (str)
            3. test_paths (List[str])

            Each path (str) should be a csv file containing the following columns:

            ====================  ====================
            column                description
            ====================  ====================
            id                    (str) - the unique id for this data point
            wav_path              (str) - the absolute path of the waveform file
            label                 (str) - a string label of the waveform
            start_sec             (float) - optional, load the waveform from :code:`start_sec` seconds. If not presented or is :code:`math.nan`, load from the beginning.
            end_sec               (float) - optional, load the waveform from :code:`end_sec` seconds. If not presented or is :code:`math.nan`, load to the end.
            ====================  ====================
        """
        return iemocap_for_superb(
            **self._get_current_arguments(flatten_dict="prepare_data")
        )



from s3prl.dataio.dataset.frame_label import (
    chunk_labels_to_frame_tensor_label,
    chunking,
)


def test_chunking():
    chunks = list(chunking(0.0, 8.5, 2.0, 1.0, False))
    assert len(chunks) == 7

    chunks = list(chunking(1.1, 8.5, 2.0, 1.0, True))
    assert len(chunks) == 8


def test_frame_tensor_label():
    labels = [
        (0, 3.0, 4.1),
        (1, 1.2, 3.2),
    ]
    label = chunk_labels_to_frame_tensor_label(1.5, 4.0, labels, 3, 160)
    assert label[-1, 0] == 1
    assert label[0, 1] == 1



import pytest
from dotenv import dotenv_values

from s3prl.dataio.corpus.librilight import LibriLight
from s3prl.dataio.corpus.librispeech import LibriSpeech

libri_stats = {
    "train-clean-100": 28539,
    "train-clean-360": 104014,
    "train-other-500": 148688,
    "dev-clean": 2703,
    "dev-other": 2864,
    "test-clean": 2620,
    "test-other": 2939,
}


@pytest.mark.corpus
def test_librispeech_dataset():
    config = dotenv_values()
    dataset_root = config["LibriSpeech"]
    dataset = LibriSpeech(
        dataset_root,
        train_split=[
            "train-clean-100",
            "train-clean-360",
        ],  # FIXME (Leo): I temporary do not have space for train-other-500 ...
        valid_split=["dev-clean", "dev-other"],
        test_split=["test-clean", "test-other"],
    )
    data = dataset.all_data
    assert len(data) == 292367 - libri_stats["train-other-500"]


@pytest.mark.corpus
def test_librilight():
    config = dotenv_values()
    train_corpus = LibriLight(config["LibriLight"])
    eval_corpus = LibriSpeech(config["LibriSpeech"], 4, [])

    train_data = train_corpus.all_data
    _, valid_data, test_data = eval_corpus.data_split

    assert len(train_data) == 48



from dataclasses import dataclass

from s3prl.upstream.apc.audio import create_transform as apc_feat_extracter
from s3prl.upstream.baseline.extracter import get_extracter as kaldi_feat_extracter
from s3prl.upstream.baseline.preprocessor import (
    get_preprocessor as online_feat_extracter,
)
from s3prl.upstream.npc.audio import create_transform as npc_feat_extracter

from .base import AugmentedDynamicItemDataset, DataPipe


@dataclass
class ExtractKaldiFeat(DataPipe):
    kaldi: dict = None
    delta: dict = None
    cmvn: dict = None
    wav_name: str = "wav"
    feat_name: str = "feat"
    """
    Args:
        kaldi (dict): args for the kaldi extracter
        delta (dict): args for applying delta on features
        cmvn (dict): args for applying cmvn on features
        wav_name (str): handle for the `takes` (input)
        feat_name (str): handle for the `provides` (output)
    """

    def extract_feat(self, extracter, wav):
        """
        (wav_seq_len, 1) -> extracter -> (feat_seq_len, feat_dim)
        """
        feat = extracter(wav)
        return feat

    def __call__(self, dataset: AugmentedDynamicItemDataset):
        _audio_config = {
            "kaldi": self.kaldi,
            "delta": self.delta,
            "cmvn": self.cmvn,
        }
        extracter, feat_dim, frame_shift = kaldi_feat_extracter(_audio_config)
        dataset.add_tool("extracter", extracter)
        dataset.add_tool("feat_dim", feat_dim)
        dataset.add_tool("frame_shift", frame_shift)
        dataset.add_dynamic_item(
            self.extract_feat,
            takes=["extracter", self.wav_name],
            provides=self.feat_name,
        )
        return dataset


@dataclass
class ExtractOnlineFeat(DataPipe):
    win_ms: int = 25
    hop_ms: int = 10
    n_freq: int = 201
    n_mels: int = 80
    n_mfcc: int = 13
    input: dict = None
    target: dict = None
    wav_name: str = "wav"
    feat_name: str = "feat"
    """
    Args:
        win_ms (int): window size in ms
        hop_ms (int): hop size in ms
        n_freq (int): number of frequency bins
        n_mels (int): number of mel features
        n_mfcc (int): number of mfcc features
        input (dict): args for the input feat, example - {"channel": 0, "cmvn": True, "delta": 0, "feat_type": "mel", "log": True,}
        target (dict): args for the output feat, example - {"channel": 1, "cmvn": True, "delta": 0, "feat_type": "mel", "log": True,}
        wav_name (str): handle for the `takes` (input)
        feat_name (str): handle for the `provides` (output)
    """

    def extract_feat(self, extracter, wav):
        """
        (wav_seq_len, 1) -> permute + unsqueeze ->
        (1, 1, wav_seq_len) -> extracter -> (feat_seq_len, feat_dim)
        """
        wav = wav.permute(1, 0).unsqueeze(0)
        feat = extracter(wav)[0][0]
        return feat

    def __call__(self, dataset: AugmentedDynamicItemDataset):
        _audio_config = {
            "win_ms": self.win_ms,
            "hop_ms": self.hop_ms,
            "n_freq": self.n_freq,
            "n_mels": self.n_mels,
            "n_mfcc": self.n_mfcc,
            "input": self.input,
            "target": self.target,
        }
        extracter, feat_dim, _ = online_feat_extracter(_audio_config)
        dataset.add_tool("extracter", extracter)
        dataset.add_tool("feat_dim", feat_dim)
        dataset.add_dynamic_item(
            self.extract_feat,
            takes=["extracter", self.wav_name],
            provides=self.feat_name,
        )
        return dataset


@dataclass
class ExtractApcFeat(DataPipe):
    feat_type: str = "fbank"
    feat_dim: int = 80
    frame_length: int = 25
    frame_shift: int = 10
    decode_wav: bool = False
    cmvn: bool = True
    wav_name: str = "wav"
    feat_name: str = "feat"
    """
    Args:
        feat_type (str): feature type
        feat_dim (int): feature dimension
        frame_length (int): window size in ms
        frame_shift (int): hop size in ms
        decode_wav (bool): whether to decode wav
        cmvn (bool): whether to apply uttr.-wised CMVN on feature
        wav_name (str): handle for the `takes` (input)
        feat_name (str): handle for the `provides` (output)
    """

    def extract_feat(self, extracter, wav):
        """
        (wav_seq_len, 1) -> permute ->
        (1, wav_seq_len) -> extracter -> (feat_seq_len, feat_dim)
        """
        feat = extracter(wav.permute(1, 0))
        return feat

    def __call__(self, dataset: AugmentedDynamicItemDataset):
        _audio_config = {
            "feat_type": self.feat_type,
            "feat_dim": self.feat_dim,
            "frame_length": self.frame_length,
            "frame_shift": self.frame_shift,
            "decode_wav": self.decode_wav,
            "cmvn": self.cmvn,
        }
        extracter, feat_dim = apc_feat_extracter(_audio_config)
        dataset.add_tool("extracter", extracter)
        dataset.add_tool("feat_dim", feat_dim)
        dataset.add_dynamic_item(
            self.extract_feat,
            takes=["extracter", self.wav_name],
            provides=self.feat_name,
        )
        return dataset


@dataclass
class ExtractNpcFeat(DataPipe):
    feat_type: str = "fbank"
    feat_dim: int = 80
    frame_length: int = 25
    frame_shift: int = 10
    decode_wav: bool = False
    cmvn: bool = True
    wav_name: str = "wav"
    feat_name: str = "feat"
    """
    Args:
        feat_type (str): feature type
        feat_dim (int): feature dimension
        frame_length (int): window size in ms
        frame_shift (int): hop size in ms
        decode_wav (bool): whether to decode wav
        cmvn (bool): whether to apply uttr.-wised CMVN on feature
        wav_name (str): handle for the `takes` (input)
        feat_name (str): handle for the `provides` (output)
    """

    def extract_feat(self, extracter, wav):
        """
        (wav_seq_len, 1) -> permute ->
        (1, wav_seq_len) -> extracter -> (feat_seq_len, feat_dim)
        """
        feat = extracter(wav.permute(1, 0))
        return feat

    def __call__(self, dataset: AugmentedDynamicItemDataset):
        _audio_config = {
            "feat_type": self.feat_type,
            "feat_dim": self.feat_dim,
            "frame_length": self.frame_length,
            "frame_shift": self.frame_shift,
            "decode_wav": self.decode_wav,
            "cmvn": self.cmvn,
        }
        extracter, feat_dim = npc_feat_extracter(_audio_config)
        dataset.add_tool("extracter", extracter)
        dataset.add_tool("feat_dim", feat_dim)
        dataset.add_dynamic_item(
            self.extract_feat,
            takes=["extracter", self.wav_name],
            provides=self.feat_name,
        )
        return dataset



from pathlib import Path

import pytest
from dotenv import dotenv_values

from s3prl.dataio.corpus.voxceleb1sid import VoxCeleb1SID


@pytest.mark.corpus
@pytest.mark.parametrize("use_cache", [False, True])
def test_voxceleb1sid(use_cache):
    config = dotenv_values()
    voxceleb1 = Path(config["VoxCeleb1"])
    if voxceleb1.is_dir():
        train_data, valid_data, test_data = VoxCeleb1SID(voxceleb1).data_split
    else:
        raise ValueError("Please set the VoxCeleb1 path in .env")



from re import X

import torch.nn as nn


class CLSTM(nn.Module):
    # sample rate and embedding sizes are required model attributes for the HEAR API
    sample_rate = 16000
    embedding_size = 1024
    scene_embedding_size = embedding_size
    timestamp_embedding_size = embedding_size

    def __init__(self):
        super().__init__()
        self.conv = nn.Sequential(
            nn.Conv2d(1, 256, 9, stride=1, padding=1),
            nn.BatchNorm2d(256),
            nn.ReLU(),
            nn.MaxPool2d(2, stride=2),
            nn.Conv2d(256, 256, (4, 3), stride=1, padding=1),
            nn.BatchNorm2d(256),
            nn.ReLU(),
            nn.MaxPool2d(2, stride=2),
        )

        self.bilstm = nn.LSTM(
            input_size=3584,
            hidden_size=512,
            num_layers=1,
            bidirectional=True,
            batch_first=True,
        )
        self.fc_final = nn.Sequential(
            nn.Linear(1024, 1024),
            nn.ReLU(),
        )

    def forward(self, x):
        x = self.conv(x)  # (batch, ch, mel, time)
        x = x.permute(0, 3, 2, 1)  # (batch, time, mel, ch)
        B, T, D, C = x.shape
        x = x.reshape((B, T, C * D))  # (batch, time, mel*ch)
        self.bilstm.flatten_parameters()
        x, _ = self.bilstm(x)
        x = self.fc_final(x[:, -1, :])
        return x


# if __name__ == '__main__':
#     import torch
#     from torchinfo import summary
#     device = torch.device("cuda") # PyTorch v0.4.0
#     tensor = torch.randn(2, 1, 64, 96).to(device)
#     model = CLSTM().to(device)
#     x = model.forward(tensor)
#     print(x.shape)
# summary(model, input_size=(2, 1, 64, 96))



"""
The setting of Superb IC

Authors
  * Wei-Cheng Tseng 2021
  * Leo 2021
  * Leo 2022
"""

import logging
import pickle
from pathlib import Path

import pandas as pd
import torch
from omegaconf import MISSING
from torch.utils.data import Dataset

from s3prl.dataio.corpus.fluent_speech_commands import FluentSpeechCommands
from s3prl.dataio.dataset import EncodeCategories, LoadAudio
from s3prl.dataio.encoder.category import CategoryEncoders
from s3prl.dataio.sampler import FixedBatchSizeBatchSampler
from s3prl.nn.linear import MeanPoolingLinear
from s3prl.task.utterance_classification_task import (
    UtteranceMultiClassClassificationTask,
)

from .run import Common

logger = logging.getLogger(__name__)


__all__ = [
    "fsc_for_multi_classification",
    "SuperbIC",
]


def fsc_for_multi_classification(
    target_dir: str,
    cache_dir: str,
    dataset_root: str,
    n_jobs: int = 6,
    get_path_only: bool = False,
):
    """
    Prepare Fluent Speech Command for multi-class classfication
    following :obj:`SuperbIC.prepare_data` format. The standard usage
    is to use three labels jointly: action, object, and location.

    Args:
        dataset_root (str): The root path of Fluent Speech Command
        n_jobs (int): to speed up the corpus parsing procedure
    """
    target_dir = Path(target_dir)

    train_path = target_dir / f"train.csv"
    valid_path = target_dir / f"valid.csv"
    test_paths = [target_dir / f"test.csv"]

    if get_path_only:
        return train_path, valid_path, test_paths

    def format_fields(data_points: dict):
        return {
            key: dict(
                wav_path=value["path"],
                labels=f"{value['action']} ; {value['object']} ; {value['location']}",
            )
            for key, value in data_points.items()
        }

    corpus = FluentSpeechCommands(dataset_root, n_jobs)
    train_data, valid_data, test_data = corpus.data_split
    train_data = format_fields(train_data)
    valid_data = format_fields(valid_data)
    test_data = format_fields(test_data)

    def dict_to_csv(data_dict, csv_path):
        keys = sorted(list(data_dict.keys()))
        fields = sorted(data_dict[keys[0]].keys())
        data = dict()
        for field in fields:
            data[field] = []
            for key in keys:
                data[field].append(data_dict[key][field])
        data["id"] = keys
        df = pd.DataFrame(data)
        df.to_csv(csv_path, index=False)

    dict_to_csv(train_data, train_path)
    dict_to_csv(valid_data, valid_path)
    dict_to_csv(test_data, test_paths[0])

    return train_path, valid_path, test_paths


class SuperbIC(Common):
    def default_config(self) -> dict:
        return dict(
            start=0,
            stop=None,
            target_dir=MISSING,
            cache_dir=None,
            remove_all_cache=False,
            prepare_data=dict(
                dataset_root=MISSING,
            ),
            build_encoder=dict(),
            build_dataset=dict(),
            build_batch_sampler=dict(
                train=dict(
                    batch_size=32,
                    shuffle=True,
                ),
                valid=dict(
                    batch_size=32,
                ),
                test=dict(
                    batch_size=32,
                ),
            ),
            build_upstream=dict(
                name=MISSING,
            ),
            build_featurizer=dict(
                layer_selections=None,
                normalize=False,
            ),
            build_downstream=dict(
                hidden_size=256,
            ),
            build_model=dict(
                upstream_trainable=False,
            ),
            build_task=dict(),
            build_optimizer=dict(
                name="Adam",
                conf=dict(
                    lr=1.0e-4,
                ),
            ),
            build_scheduler=dict(
                name="ExponentialLR",
                gamma=0.9,
            ),
            save_model=dict(),
            save_task=dict(),
            train=dict(
                total_steps=200000,
                log_step=100,
                eval_step=5000,
                save_step=250,
                gradient_clipping=1.0,
                gradient_accumulate=1,
                valid_metric="accuracy",
                valid_higher_better=True,
                auto_resume=True,
                resume_ckpt_dir=None,
            ),
        )

    def prepare_data(
        self,
        prepare_data: dict,
        target_dir: str,
        cache_dir: str,
        get_path_only: bool = False,
    ):
        """
        Prepare the task-specific data metadata (path, labels...).
        By default call :obj:`fsc_for_multi_classification` with :code:`**prepare_data`

        Args:
            prepare_data (dict): same in :obj:`default_config`,
                arguments for :obj:`fsc_for_multi_classification`
            target_dir (str): Parse your corpus and save the csv file into this directory
            cache_dir (str): If the parsing or preprocessing takes too long time, you can save
                the temporary files into this directory. This directory is expected to be shared
                across different training sessions (different hypers and :code:`target_dir`)
            get_path_only (str): Directly return the filepaths no matter they exist or not.

        Returns:
            tuple

            1. train_path (str)
            2. valid_path (str)
            3. test_paths (List[str])

            Each path (str) should be a csv file containing the following columns:

            ====================  ====================
            column                description
            ====================  ====================
            id                    (str) - the unique id for this data point
            wav_path              (str) - the absolute path of the waveform file
            labels                (str) - the string labels of the waveform, separated by a ';'
            ====================  ====================

            The number of the label columns can be arbitrary.
        """
        return fsc_for_multi_classification(
            **self._get_current_arguments(flatten_dict="prepare_data")
        )

    def build_encoder(
        self,
        build_encoder: dict,
        target_dir: str,
        cache_dir: str,
        train_csv_path: str,
        valid_csv_path: str,
        test_csv_paths: list,
        get_path_only: bool = False,
    ):
        """
        Build the encoder (for the labels) given the data metadata, and return the saved encoder path.
        By default generate and save a :obj:`s3prl.dataio.encoder.CategoryEncoders` from all the columns
        prefixing :code:`label` from all the csv files.

        Args:
            build_encoder (dict): same in :obj:`default_config`, no argument supported for now
            target_dir (str): Save your encoder into this directory
            cache_dir (str): If the preprocessing takes too long time, you can save
                the temporary files into this directory. This directory is expected to be shared
                across different training sessions (different hypers and :code:`target_dir`)
            train_csv_path (str): the train path from :obj:`prepare_data`
            valid_csv_path (str): the valid path from :obj:`prepare_data`
            test_csv_paths (List[str]): the test paths from :obj:`prepare_data`
            get_path_only (bool): Directly return the filepaths no matter they exist or not.

        Returns:
            str

            tokenizer_path: The tokenizer should be saved in the pickle format
        """
        encoder_path = Path(target_dir) / "encoder.pkl"
        if get_path_only:
            return encoder_path

        train_csv = pd.read_csv(train_csv_path)
        valid_csv = pd.read_csv(valid_csv_path)
        test_csvs = [pd.read_csv(path) for path in test_csv_paths]
        all_csv = pd.concat([train_csv, valid_csv, *test_csvs])

        multilabels = [
            [label.strip() for label in multilabel.split(";")]
            for multilabel in all_csv["labels"].tolist()
        ]
        encoder = CategoryEncoders(
            [single_category_labels for single_category_labels in zip(*multilabels)]
        )
        with open(encoder_path, "wb") as f:
            pickle.dump(encoder, f)

        return encoder

    def build_dataset(
        self,
        build_dataset: dict,
        target_dir: str,
        cache_dir: str,
        mode: str,
        data_csv: str,
        encoder_path: str,
        frame_shift: int,
    ):
        """
        Build the dataset for train/valid/test.

        Args:
            build_dataset (dict): same in :obj:`default_config`, no argument supported for now
            target_dir (str): Current experiment directory
            cache_dir (str): If the preprocessing takes too long time, you can save
                the temporary files into this directory. This directory is expected to be shared
                across different training sessions (different hypers and :code:`target_dir`)
            mode (str): train/valid/test
            data_csv (str): The metadata csv file for the specific :code:`mode`
            encoder_path (str): The pickled encoder path for encoding the labels

        Returns:
            torch Dataset

            For all train/valid/test mode, the dataset should return each item as a dictionary
            containing the following keys:

            ====================  ====================
            key                   description
            ====================  ====================
            x                     (torch.FloatTensor) - the waveform in (seq_len, 1)
            x_len                 (int) - the waveform length :code:`seq_len`
            class_ids             (torch.LongTensor) - the encoded class ids. shape: (num_class, )
            labels                (List[str]) - the class name. length: num_class
            unique_name           (str) - the unique id for this datapoint
            ====================  ====================
        """
        csv = pd.read_csv(data_csv)
        ids = csv["id"].tolist()

        audio_loader = LoadAudio(csv["wav_path"].tolist())

        with open(encoder_path, "rb") as f:
            encoder = pickle.load(f)

        label_encoder = EncodeCategories(
            [
                [label.strip() for label in multilabel.split(";")]
                for multilabel in csv["labels"].tolist()
            ],
            encoder,
        )

        class Dataset:
            def __len__(self):
                return len(audio_loader)

            def __getitem__(self, index: int):
                audio = audio_loader[index]
                label = label_encoder[index]
                return {
                    "x": audio["wav"],
                    "x_len": audio["wav_len"],
                    "class_ids": label["class_ids"],
                    "labels": label["labels"],
                    "unique_name": ids[index],
                }

        dataset = Dataset()
        return dataset

    def build_batch_sampler(
        self,
        build_batch_sampler: dict,
        target_dir: str,
        cache_dir: str,
        mode: str,
        data_csv: str,
        dataset: Dataset,
    ):
        """
        Return the batch sampler for torch DataLoader.
        By default call :obj:`superb_sid_batch_sampler` with :code:`**build_batch_sampler`.

        Args:
            build_batch_sampler (dict): same in :obj:`default_config`

                ====================  ====================
                key                   description
                ====================  ====================
                train                 (dict) - arguments for :obj:`FixedBatchSizeBatchSampler`
                valid                 (dict) - arguments for :obj:`FixedBatchSizeBatchSampler`
                test                  (dict) - arguments for :obj:`FixedBatchSizeBatchSampler`
                ====================  ====================

            target_dir (str): Current experiment directory
            cache_dir (str): If the preprocessing takes too long time, save
                the temporary files into this directory. This directory is expected to be shared
                across different training sessions (different hypers and :code:`target_dir`)
            mode (str): train/valid/test
            data_csv (str): the :code:`mode` specific csv from :obj:`prepare_data`
            dataset: the dataset from :obj:`build_dataset`

        Returns:
            batch sampler for torch DataLoader
        """

        def _build_batch_sampler(
            train: dict = None, valid: dict = None, test: dict = None
        ):
            if mode == "train":
                return FixedBatchSizeBatchSampler(dataset, **train)
            elif mode == "valid":
                return FixedBatchSizeBatchSampler(dataset, **valid)
            elif mode == "test":
                return FixedBatchSizeBatchSampler(dataset, **test)

        return _build_batch_sampler(**build_batch_sampler)

    def build_downstream(
        self,
        build_downstream: dict,
        downstream_input_size: int,
        downstream_output_size: int,
        downstream_input_stride: int,
    ):
        """
        Return the task-specific downstream model.
        By default build the :obj:`MeanPoolingLinear` model

        Args:
            build_downstream (dict): same in :obj:`default_config`,
                support arguments of :obj:`MeanPoolingLinear`
            downstream_input_size (int): the required input size of the model
            downstream_output_size (int): the required output size of the model
            downstream_input_stride (int): the input feature's stride (from 16 KHz)

        Returns:
            :obj:`AbsUtteranceModel`
        """
        model = MeanPoolingLinear(
            downstream_input_size, downstream_output_size, **build_downstream
        )
        return model

    def build_task(
        self,
        build_task: dict,
        model: torch.nn.Module,
        encoder,
        valid_df: pd.DataFrame = None,
        test_df: pd.DataFrame = None,
    ):
        """
        Build the task, which defines the logics for every train/valid/test forward step for the :code:`model`,
        and the logics for how to reduce all the batch results from multiple train/valid/test steps into metrics

        By default build :obj:`UtteranceMultiClassClassificationTask`

        Args:
            build_task (dict): same in :obj:`default_config`, no argument supported for now
            model (torch.nn.Module): the model built by :obj:`build_model`
            encoder: the encoder built by :obj:`build_encoder`
            valid_df (pd.DataFrame): metadata of the valid set
            test_df (pd.DataFrame): metadata of the test set

        Returns:
            Task
        """
        return UtteranceMultiClassClassificationTask(model, encoder)



"""
Authors:
    - Leo (2022)
"""

from typing import Any, List, Tuple

import pandas as pd
import torch

from .base import Dataset
from .load_audio import LoadAudio

__all__ = [
    "chunking",
    "scale_labels_secs",
    "get_chunk_labels",
    "chunk_labels_to_frame_tensor_label",
    "FrameLabelDataset",
]


def chunking(
    start_sec: float,
    end_sec: float,
    chunk_secs: float,
    step_secs: float,
    use_unfull_chunks: bool = True,
) -> List[Tuple[float, float]]:
    """
    Produce chunks (start, end points) from a given start, end seconds

    Args:
        start_sec (float): The start second of the utterance
        end_sec (float): The end second of the utterance
        chunk_secs (float): The length (in seconds) of a chunked chunk
        step_secs (float): The stride seconds between chunks
        use_unfull_chunks (bool): Whether to produce chunks shorter than :code:`chunk_secs`
            at the end of the recording

    Returns:
        List[Tuple[float, float]]: Each tuple describes the starting point (in sec)
            and the ending point (in sec) of each chunk in order
    """

    start, end = start_sec, end_sec
    while end - start > 0:
        if end - start >= chunk_secs:
            yield start, start + chunk_secs

        elif use_unfull_chunks:
            yield start, end

        start = start + step_secs


def scale_labels_secs(labels: List[Tuple[Any, float, float]], ratio: float):
    """
    When the recording length is changed due to like pitch or speed manipulation,
    the start/end timestamp (in seconds) should also be changed

    Args:
        labels (List[Tuple[Any, float, float]]): each chunk label is in (label, start_sec, end_sec)
        ratio (float): the scaling ratio

    Returns:
        List[Tuple[Any, float, float]]: the scaled labels
    """
    assert ratio > 0
    return [(label, start * ratio, end * ratio) for label, start, end in labels]


def get_chunk_labels(
    start_sec: float,
    end_sec: float,
    labels: List[Tuple[Any, float, float]],
):
    """
    Given a pair a start, end points, filter out the relevant labels from the given :code:`labels`
    and refine the start/end points of each label to reside between :code:`start_sec` and :code:`end_sec`

    Args:
        start_sec (float): the starting point
        end_sec (float): the ending point
        labels (List[Tuple[Any, float, float]]): the chunk labels

    Returns:
        List[Tuple[str, float, float]]: filtered labels. Only the labels relevant to the assigned
            start/end point are left
    """

    for label, start, end in labels:
        assert start < end, f"start ({start}) >= end ({end})"
        if start >= end_sec:
            continue
        if end <= start_sec:
            continue
        yield label, max(start_sec, start), min(end_sec, end)


def chunk_labels_to_frame_tensor_label(
    start_sec: float,
    end_sec: float,
    labels: List[Tuple[int, float, float]],
    num_class: int,
    frame_shift: int,
    sample_rate: int = 16000,
):
    """
    Produce frame-level labels for the given chunk labels

    Args:
        start_sec (float): the starting point of the chunk
        end_sec (float): the ending point of the chunk
        labels (List[Tuple[int, float, float]]): the chunk labels, each label is a tuple
            in (class_id, start_sec, end_sec)
        num_class (int): number of classes
        frame_shift (int): produce a frame per :code:`frame_shift` samples
        sample_rate (int): the sample rate of the recording. default: 16000

    Returns:
        torch.FloatTensor: shape (num_frames, num_class).
            the binary frame labels for the given :code:`labels`
    """
    labels = get_chunk_labels(start_sec, end_sec, labels)

    duration = end_sec - start_sec
    num_frames = len(range(0, round(duration * sample_rate), frame_shift))

    frame_labels = torch.zeros(num_frames, num_class)
    for class_id, start, end in labels:
        assert start >= start_sec, f"{start} < {start_sec}"
        assert end >= start_sec, f"{end} < {start_sec}"

        start_frame = round((start - start_sec) * sample_rate) // frame_shift
        end_frame = round((end - start_sec) * sample_rate) // frame_shift
        frame_labels[start_frame : end_frame + 1, class_id] = 1.0

    return frame_labels


class FrameLabelDataset(Dataset):
    """
    Args:
        df (pd.DataFrame): the dataframe should have the following columns
            record_id (str), wav_path (str), duration (float), utt_id (str),
            label (int), start_sec (float), end_sec (float)
    """

    def __init__(
        self,
        df: pd.DataFrame,
        num_class: int,
        frame_shift: int,
        chunk_secs: float,
        step_secs: float,
        use_unfull_chunks: bool = True,
        load_audio_conf: dict = None,
        sample_rate: int = 16000,
    ) -> None:
        super().__init__()

        self.num_class = num_class
        self.frame_shift = frame_shift
        self.sample_rate = sample_rate

        recording_df = df[["record_id", "wav_path", "duration"]].drop_duplicates()
        record_ids = recording_df["record_id"].tolist()
        record_to_labels = {}
        for record_id in record_ids:
            subset_df = df[df["record_id"] == record_id]
            labels = list(
                zip(subset_df["label"], subset_df["start_sec"], subset_df["end_sec"])
            )
            record_to_labels[record_id] = labels

        self.chunked_utts = []
        for _, row in recording_df.iterrows():
            chunks = chunking(
                0.0,
                row["duration"],
                chunk_secs,
                step_secs,
                use_unfull_chunks,
            )
            for chunk_id, (start, end) in enumerate(chunks):
                labels = list(
                    get_chunk_labels(start, end, record_to_labels[row["record_id"]])
                )
                self.chunked_utts.append(
                    {
                        "record_id": row["record_id"],
                        "chunk_id": chunk_id,
                        "wav_path": row["wav_path"],
                        "start_sec": start,
                        "end_sec": end,
                        "unique_name": f"{row['record_id']}-{start}-{end}",
                        "labels": labels,
                    }
                )

        def flatten(data: List[dict], key: str):
            return [item[key] for item in data]

        self.audio_loader = LoadAudio(
            flatten(self.chunked_utts, "wav_path"),
            flatten(self.chunked_utts, "start_sec"),
            flatten(self.chunked_utts, "end_sec"),
            **(load_audio_conf or {}),
        )

    def __len__(self) -> int:
        return len(self.chunked_utts)

    def getinfo(self, index: int):
        return self.chunked_utts[index]

    def __getitem__(self, index):
        info = self.getinfo(index)
        audio = self.audio_loader[index]
        label = chunk_labels_to_frame_tensor_label(
            info["start_sec"],
            info["end_sec"],
            info["labels"],
            self.num_class,
            self.frame_shift,
            self.sample_rate,
        )
        return {
            "x": audio["wav"],
            "x_len": audio["wav_len"],
            "y": label,
            "y_len": len(label),
            "unique_name": info["unique_name"],
            "labels": info["labels"],
            "record_id": info["record_id"],
            "chunk_id": info["chunk_id"],
        }



from s3prl.dataio.encoder.tokenizer import CharacterTokenizer, default_phoneme_tokenizer


def test_tokenizer():
    char_tokenizer = CharacterTokenizer()
    phone_tokenizer = default_phoneme_tokenizer()

    char_text = "HELLO WORLD"
    char_text_enc = char_tokenizer.encode(char_text)
    char_text_dec = char_tokenizer.decode(char_text_enc)

    assert isinstance(char_text_enc, list)
    assert char_text == char_text_dec



# Copyright Hear Benchmark Team

from collections import ChainMap
from functools import partial
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import numpy as np
import pandas as pd
import sed_eval
import torch

# Can we get away with not using DCase for every event-based evaluation??
from dcase_util.containers import MetaDataContainer
from scipy import stats
from sklearn.metrics import average_precision_score, roc_auc_score


def label_vocab_as_dict(df: pd.DataFrame, key: str, value: str) -> Dict:
    """
    Returns a dictionary of the label vocabulary mapping the label column to
    the idx column. key sets whether the label or idx is the key in the dict. The
    other column will be the value.
    """
    if key == "label":
        # Make sure the key is a string
        df["label"] = df["label"].astype(str)
        value = "idx"
    else:
        assert key == "idx", "key argument must be either 'label' or 'idx'"
        value = "label"
    return df.set_index(key).to_dict()[value]


def label_to_binary_vector(label: List, num_labels: int) -> torch.Tensor:
    """
    Converts a list of labels into a binary vector
    Args:
        label: list of integer labels
        num_labels: total number of labels

    Returns:
        A float Tensor that is multi-hot binary vector
    """
    # Lame special case for multilabel with no labels
    if len(label) == 0:
        # BCEWithLogitsLoss wants float not long targets
        binary_labels = torch.zeros((num_labels,), dtype=torch.float)
    else:
        binary_labels = torch.zeros((num_labels,)).scatter(0, torch.tensor(label), 1.0)

    # Validate the binary vector we just created
    assert set(torch.where(binary_labels == 1.0)[0].numpy()) == set(label)
    return binary_labels


def validate_score_return_type(ret: Union[Tuple[Tuple[str, float], ...], float]):
    """
    Valid return types for the metric are
        - tuple(tuple(string: name of the subtype, float: the value)): This is the
            case with sed eval metrics. They can return (("f_measure", value),
            ("precision", value), ...), depending on the scores
            the metric should is supposed to return. This is set as `scores`
            attribute in the metric.
        - float: Standard metric behaviour

    The downstream prediction pipeline is able to handle these two types.
    In case of the tuple return type, the value of the first entry in the
    tuple will be used as an optimisation criterion wherever required.
    For instance, if the return is (("f_measure", value), ("precision", value)),
    the value corresponding to the f_measure will be used ( for instance in
    early stopping if this metric is the primary score for the task )
    """
    if isinstance(ret, tuple):
        assert all(
            type(s) == tuple and type(s[0]) == str and type(s[1]) == float for s in ret
        ), (
            "If the return type of the score is a tuple, all the elements "
            "in the tuple should be tuple of type (string, float)"
        )
    elif isinstance(ret, float):
        pass
    else:
        raise ValueError(
            f"Return type {type(ret)} is unexpected. Return type of "
            "the score function should either be a "
            "tuple(tuple) or float. "
        )


class ScoreFunction:
    """
    A simple abstract base class for score functions
    """

    # TODO: Remove label_to_idx?
    def __init__(
        self,
        label_to_idx: Dict[str, int],
        name: Optional[str] = None,
        maximize: bool = True,
    ):
        """
        :param label_to_idx: Map from label string to integer index.
        :param name: Override the name of this scoring function.
        :param maximize: Maximize this score? (Otherwise, it's a loss or energy
            we want to minimize, and I guess technically isn't a score.)
        """
        self.label_to_idx = label_to_idx
        if name:
            self.name = name
        self.maximize = maximize

    def __call__(self, *args, **kwargs) -> Union[Tuple[Tuple[str, float], ...], float]:
        """
        Calls the compute function of the metric, and after validating the output,
        returns the metric score
        """
        ret = self._compute(*args, **kwargs)
        validate_score_return_type(ret)
        return ret

    def _compute(
        self, predictions: Any, targets: Any, **kwargs
    ) -> Union[Tuple[Tuple[str, float], ...], float]:
        """
        Compute the score based on the predictions and targets.
        This is a private function and the metric should be used as a functor
        by calling the `__call__` method which calls this and also validates
        the return type
        """
        raise NotImplementedError("Inheriting classes must implement this function")

    def __str__(self):
        return self.name


class Top1Accuracy(ScoreFunction):
    name = "top1_acc"

    def _compute(self, predictions: np.ndarray, targets: np.ndarray, **kwargs) -> float:
        assert predictions.ndim == 2
        assert targets.ndim == 2  # One hot
        # Compute the number of correct predictions
        correct = 0
        for target, prediction in zip(targets, predictions):
            assert prediction.ndim == 1
            assert target.ndim == 1
            predicted_class = np.argmax(prediction)
            target_class = np.argmax(target)

            if predicted_class == target_class:
                correct += 1

        return correct / len(targets)


class ChromaAccuracy(ScoreFunction):
    """
    Score specifically for pitch detection -- converts all pitches to chroma first.
    This score ignores octave errors in pitch classification.
    """

    name = "chroma_acc"

    def _compute(self, predictions: np.ndarray, targets: np.ndarray, **kwargs) -> float:
        # Compute the number of correct predictions
        correct = 0
        for target, prediction in zip(targets, predictions):
            assert prediction.ndim == 1
            assert target.ndim == 1
            predicted_class = np.argmax(prediction)
            target_class = np.argmax(target)

            # Ignore octave errors by converting the predicted class to chroma before
            # checking for correctness.
            if predicted_class % 12 == target_class % 12:
                correct += 1

        return correct / len(targets)


class SoundEventScore(ScoreFunction):
    """
    Scores for sound event detection tasks using sed_eval
    """

    # Score class must be defined in inheriting classes
    score_class: sed_eval.sound_event.SoundEventMetrics = None

    def __init__(
        self,
        label_to_idx: Dict[str, int],
        scores: Tuple[str],
        params: Dict = None,
        name: Optional[str] = None,
        maximize: bool = True,
    ):
        """
        :param scores: Scores to use, from the list of overall SED eval scores.
            The first score in the tuple will be the primary score for this metric
        :param params: Parameters to pass to the scoring function,
                       see inheriting children for details.
        """
        if params is None:
            params = {}
        super().__init__(label_to_idx=label_to_idx, name=name, maximize=maximize)
        self.scores = scores
        self.params = params
        assert self.score_class is not None

    def _compute(
        self, predictions: Dict, targets: Dict, **kwargs
    ) -> Tuple[Tuple[str, float], ...]:
        # Containers of events for sed_eval
        reference_event_list = self.sed_eval_event_container(targets)
        estimated_event_list = self.sed_eval_event_container(predictions)

        # This will break in Python < 3.6 if the dict order is not
        # the insertion order I think. I'm a little worried about this line
        scores = self.score_class(
            event_label_list=list(self.label_to_idx.keys()), **self.params
        )

        for filename in predictions:
            scores.evaluate(
                reference_event_list=reference_event_list.filter(filename=filename),
                estimated_event_list=estimated_event_list.filter(filename=filename),
            )

        # results_overall_metrics return a pretty large nested selection of scores,
        # with dicts of scores keyed on the type of scores, like f_measure, error_rate,
        # accuracy
        nested_overall_scores: Dict[
            str, Dict[str, float]
        ] = scores.results_overall_metrics()
        # Open up nested overall scores
        overall_scores: Dict[str, float] = dict(
            ChainMap(*nested_overall_scores.values())
        )
        # Return the required scores as tuples. The scores are returned in the
        # order they are passed in the `scores` argument
        return tuple([(score, overall_scores[score]) for score in self.scores])

    @staticmethod
    def sed_eval_event_container(
        x: Dict[str, List[Dict[str, Any]]]
    ) -> MetaDataContainer:
        # Reformat event list for sed_eval
        reference_events = []
        for filename, event_list in x.items():
            for event in event_list:
                reference_events.append(
                    {
                        # Convert from ms to seconds for sed_eval
                        "event_label": str(event["label"]),
                        "event_onset": event["start"] / 1000.0,
                        "event_offset": event["end"] / 1000.0,
                        "file": filename,
                    }
                )
        return MetaDataContainer(reference_events)


class SegmentBasedScore(SoundEventScore):
    """
    segment-based scores - the ground truth and system output are compared in a
    fixed time grid; sound events are marked as active or inactive in each segment;

    See https://tut-arg.github.io/sed_eval/sound_event.html#sed_eval.sound_event.SegmentBasedMetrics # noqa: E501
    for params.
    """

    score_class = sed_eval.sound_event.SegmentBasedMetrics


class EventBasedScore(SoundEventScore):
    """
    event-based scores - the ground truth and system output are compared at
    event instance level;

    See https://tut-arg.github.io/sed_eval/generated/sed_eval.sound_event.EventBasedMetrics.html # noqa: E501
    for params.
    """

    score_class = sed_eval.sound_event.EventBasedMetrics


class MeanAveragePrecision(ScoreFunction):
    """
    Average Precision is calculated in macro mode which calculates
    AP at a class level followed by macro-averaging across the classes.
    """

    name = "mAP"

    def _compute(self, predictions: np.ndarray, targets: np.ndarray, **kwargs) -> float:
        assert predictions.ndim == 2
        assert targets.ndim == 2  # One hot

        """
        Based on suggestions from Eduardo Fonseca -
        Equal weighting is assigned to each class regardless
        of its prior, which is commonly referred to as macro
        averaging, following Hershey et al. (2017); Gemmeke et al.
        (2017).
        This means that rare classes are as important as common
        classes.

        Issue with average_precision_score, when all ground truths are negative
        https://github.com/scikit-learn/scikit-learn/issues/8245
        This might come up in small tasks, where few samples are available
        """
        return average_precision_score(targets, predictions, average="macro")


class DPrime(ScoreFunction):
    """
    DPrime is calculated per class followed by averaging across the classes

    Code adapted from code provided by Eduoard Fonseca.
    """

    name = "d_prime"

    def _compute(self, predictions: np.ndarray, targets: np.ndarray, **kwargs) -> float:
        assert predictions.ndim == 2
        assert targets.ndim == 2  # One hot
        # ROC-AUC Requires more than one example for each class
        # This might fail for data in small instances, so putting this in try except
        try:
            auc = roc_auc_score(targets, predictions, average=None)

            d_prime = stats.norm().ppf(auc) * np.sqrt(2.0)
            # Calculate macro score by averaging over the classes,
            # see `MeanAveragePrecision` for reasons
            d_prime_macro = np.mean(d_prime)
            return d_prime_macro
        except ValueError:
            return np.nan


class AUCROC(ScoreFunction):
    """
    AUCROC (macro mode) is calculated per class followed by averaging across the
    classes
    """

    name = "aucroc"

    def _compute(self, predictions: np.ndarray, targets: np.ndarray, **kwargs) -> float:
        assert predictions.ndim == 2
        assert targets.ndim == 2  # One hot
        # ROC-AUC Requires more than one example for each class
        # This might fail for data in small instances, so putting this in try except
        try:
            # Macro mode auc-roc. Please check `MeanAveragePrecision`
            # for the reasoning behind using using macro mode
            auc = roc_auc_score(targets, predictions, average="macro")
            return auc
        except ValueError:
            return np.nan


available_scores: Dict[str, Callable] = {
    "top1_acc": Top1Accuracy,
    "pitch_acc": partial(Top1Accuracy, name="pitch_acc"),
    "chroma_acc": ChromaAccuracy,
    # https://tut-arg.github.io/sed_eval/generated/sed_eval.sound_event.EventBasedMetrics.html
    "event_onset_200ms_fms": partial(
        EventBasedScore,
        name="event_onset_200ms_fms",
        # If first score will be used as the primary score for this metric
        scores=("f_measure", "precision", "recall"),
        params={"evaluate_onset": True, "evaluate_offset": False, "t_collar": 0.2},
    ),
    "event_onset_50ms_fms": partial(
        EventBasedScore,
        name="event_onset_50ms_fms",
        scores=("f_measure", "precision", "recall"),
        params={"evaluate_onset": True, "evaluate_offset": False, "t_collar": 0.05},
    ),
    "event_onset_offset_50ms_20perc_fms": partial(
        EventBasedScore,
        name="event_onset_offset_50ms_20perc_fms",
        scores=("f_measure", "precision", "recall"),
        params={
            "evaluate_onset": True,
            "evaluate_offset": True,
            "t_collar": 0.05,
            "percentage_of_length": 0.2,
        },
    ),
    "segment_1s_er": partial(
        SegmentBasedScore,
        name="segment_1s_er",
        scores=("error_rate",),
        params={"time_resolution": 1.0},
        maximize=False,
    ),
    "mAP": MeanAveragePrecision,
    "d_prime": DPrime,
    "aucroc": AUCROC,
}



from typing import List

from .base import SequentialDataPipe
from .common_pipes import LoadAudio, RandomCrop, SetOutputKeys


class SpeakerVerificationPipe(SequentialDataPipe):
    """
    each item in the input dataset should have:
        wav_path: str
        label: str
    """

    def __init__(
        self,
        audio_sample_rate: int = 16000,
        audio_channel_reduction: str = "first",
        random_crop_secs: float = -1,
        sox_effects: List[List] = None,
    ):
        pipes = [
            LoadAudio(
                audio_sample_rate=audio_sample_rate,
                audio_channel_reduction=audio_channel_reduction,
                sox_effects=sox_effects,
            ),
        ]
        output_keys = dict(
            x="wav",
            x_len="wav_len",
            label="label",
            unique_name="id",
        )

        if random_crop_secs != -1:
            pipes.append(
                RandomCrop(sample_rate=audio_sample_rate, max_secs=random_crop_secs)
            )
            output_keys["x"] = "wav_crop"
            output_keys["x_len"] = "wav_crop_len"

        pipes.append(SetOutputKeys(output_keys))
        super().__init__(*pipes)



import logging
import random
from collections import OrderedDict

from s3prl.dataio.sampler import SortedBucketingSampler, SortedSliceSampler

logger = logging.getLogger(__name__)


def test_sorted_slice_sampler():
    batch_size = 16
    max_length = 16000 * 5
    lengths = [random.randint(16000 * 3, 16000 * 8) for index in range(1000)]

    sampler = SortedSliceSampler(
        lengths,
        batch_size=batch_size,
        max_length=max_length,
    )

    for epoch in range(5):
        sampler.set_epoch(epoch)
        id2length = lengths
        for batch_ids in sampler:
            batch_lengths = [id2length[idx] for idx in batch_ids]
            assert sorted(batch_lengths, reverse=True) == batch_lengths
            if batch_lengths[0] > max_length:
                assert len(batch_lengths) == batch_size // 2

        other_batch_sizes = [
            len(batch)
            for batch in sampler
            if len(batch) not in [batch_size, batch_size // 2]
        ]
        assert len(set(other_batch_sizes)) == len(other_batch_sizes)
        assert len(sampler) == len(lengths)


def test_sorted_bucketing_sampler():
    batch_size = 16
    max_length = 16000 * 5
    lengths = [random.randint(16000 * 3, 16000 * 8) for index in range(1000)]

    sampler = SortedBucketingSampler(
        lengths,
        batch_size=batch_size,
        max_length=max_length,
        shuffle=False,
    )

    for epoch in range(5):
        sampler.set_epoch(epoch)
        id2length = lengths
        for batch_ids in sampler:
            batch_lengths = [id2length[idx] for idx in batch_ids]
            assert sorted(batch_lengths, reverse=True) == batch_lengths
            if batch_lengths[0] > max_length:
                assert len(batch_lengths) == batch_size // 2

        batch_sizes = [len(batch_indices) for batch_indices in sampler]
        other_batch_sizes = [
            batch_size
            for batch_size in batch_sizes
            if batch_size not in [batch_size, batch_size // 2]
        ]
        assert len(other_batch_sizes) <= 1
        assert len(lengths) / 16 < len(sampler) < len(lengths) / 8



"""
The backbone run procedure for Speaker Diarization

Authors:
  * Leo 2022
"""

import logging
import shutil
from dataclasses import dataclass
from pathlib import Path
from typing import List

import yaml

from s3prl.problem.base import Problem
from s3prl.task.diarization import DiarizationPIT

from .util import csv_to_kaldi_dir, kaldi_dir_to_rttm, make_rttm_and_score

logger = logging.getLogger(__name__)

__all__ = ["Diarization"]


class Diarization(Problem):
    def run(
        self,
        target_dir: str,
        cache_dir: str,
        remove_all_cache: bool = False,
        start: int = 0,
        stop: int = None,
        num_workers: int = 6,
        eval_batch: int = -1,
        device: str = "cuda",
        world_size: int = 1,
        rank: int = 0,
        test_ckpt_dir: str = None,
        num_speaker: int = 2,
        prepare_data: dict = None,
        build_dataset: dict = None,
        build_batch_sampler: dict = None,
        build_collate_fn: dict = None,
        build_upstream: dict = None,
        build_featurizer: dict = None,
        build_downstream: dict = None,
        build_model: dict = None,
        build_task: dict = None,
        build_optimizer: dict = None,
        build_scheduler: dict = None,
        save_model: dict = None,
        save_task: dict = None,
        train: dict = None,
        evaluate: dict = None,
        scoring: dict = None,
    ):
        """
        ========  ====================
        stage     description
        ========  ====================
        0         Parse the corpus and save the Kaldi-style data directory for speaker diarization
        1         Train the model
        2         Inference the prediction
        3         Score the prediction
        ========  ====================

        Args:
            target_dir (str):
                The directory that stores the script result.
            cache_dir (str):
                The directory that caches the processed data.
                Default: /home/user/.cache/s3prl/data
            remove_all_cache (bool):
                Whether to remove all the cache stored under `cache_dir`.
                Default: False
            start (int):
                The starting stage of the problem script.
                Default: 0
            stop (int):
                The stoping stage of the problem script, set `None` to reach the final stage.
                Default: None
            num_workers (int): num_workers for all the torch DataLoder
            eval_batch (int):
                During evaluation (valid or test), limit the number of batch.
                This is helpful for the fast development to check everything won't crash.
                If is -1, disable this feature and evaluate the entire epoch.
                Default: -1
            device (str):
                The device type for all torch-related operation: "cpu" or "cuda"
                Default: "cuda"
            world_size (int):
                How many processes are running this script simultaneously (in parallel).
                Usually this is just 1, however if you are runnig distributed training,
                this should be > 1.
                Default: 1
            rank (int):
                When distributed training, world_size > 1. Take :code:`world_size == 8` for
                example, this means 8 processes (8 GPUs) are runing in parallel. The script
                needs to know which process among 8 processes it is. In this case, :code:`rank`
                can range from 0~7. All the 8 processes have the same :code:`world_size` but
                different :code:`rank` (process id).
            test_ckpt_dir (str):
                Specify the checkpoint path for testing. If not, use checkpoints specified by
                :code:`test_ckpts_steps`.
            num_speaker (int):
                How many speakers per utterance
            **others:
                The other arguments like :code:`prepare_data` and :code:`build_model` are
                method specific-arguments for methods like :obj:`prepare_data` and
                :obj:`build_model`, and will not be used in the core :obj:`run` logic.
                See the specific method documentation for their supported arguments and
                meaning
        """

        yaml_path = Path(target_dir) / "configs" / f"{self._get_time_tag()}.yaml"
        yaml_path.parent.mkdir(exist_ok=True, parents=True)
        with yaml_path.open("w") as f:
            yaml.safe_dump(self._get_current_arguments(), f)

        cache_dir: str = cache_dir or Path.home() / ".cache" / "s3prl" / "data"
        prepare_data: dict = prepare_data or {}
        build_dataset: dict = build_dataset or {}
        build_batch_sampler: dict = build_batch_sampler or {}
        build_collate_fn: dict = build_collate_fn or {}
        build_upstream: dict = build_upstream or {}
        build_featurizer: dict = build_featurizer or {}
        build_downstream: dict = build_downstream or {}
        build_model: dict = build_model or {}
        build_task: dict = build_task or {}
        build_optimizer: dict = build_optimizer or {}
        build_scheduler: dict = build_scheduler or {}
        save_model: dict = save_model or {}
        save_task: dict = save_task or {}
        train: dict = train or {}
        evaluate: dict = evaluate or {}
        scoring: dict = scoring or {}

        target_dir: Path = Path(target_dir)
        target_dir.mkdir(exist_ok=True, parents=True)

        cache_dir = Path(cache_dir)
        cache_dir.mkdir(exist_ok=True, parents=True)
        if remove_all_cache:
            shutil.rmtree(cache_dir, ignore_errors=True)

        stage_id = 0
        if start <= stage_id:
            logger.info(f"Stage {stage_id}: prepare data")
            train_csv, valid_csv, test_csvs = self.prepare_data(
                prepare_data, target_dir, cache_dir, get_path_only=False
            )

        train_csv, valid_csv, test_csvs = self.prepare_data(
            prepare_data, target_dir, cache_dir, get_path_only=True
        )

        def check_fn():
            assert Path(train_csv).is_file() and Path(valid_csv).is_file()
            for test_csv in test_csvs:
                assert Path(test_csv).is_file()

        self._stage_check(stage_id, stop, check_fn)

        for csv in [train_csv, valid_csv, *test_csvs]:
            data_dir = target_dir / "kaldi_data" / Path(csv).stem
            csv_to_kaldi_dir(csv, data_dir)

        train_data = target_dir / "kaldi_data" / Path(train_csv).stem
        valid_data = target_dir / "kaldi_data" / Path(valid_csv).stem
        test_datas = [target_dir / "kaldi_data" / Path(csv).stem for csv in test_csvs]

        test_rttms = []
        for test_data in test_datas:
            logger.info(f"Prepare RTTM for {test_data}")
            test_rttm = target_dir / f"{Path(test_data).stem}.rttm"
            kaldi_dir_to_rttm(test_data, test_rttm)
            test_rttms.append(test_rttm)

        model_output_size = num_speaker
        model = self.build_model(
            build_model,
            model_output_size,
            build_upstream,
            build_featurizer,
            build_downstream,
        )
        frame_shift = model.downsample_rate

        stage_id = 1
        train_dir = target_dir / "train"
        if start <= stage_id:
            logger.info(f"Stage {stage_id}: Train Model")
            train_ds, train_bs = self._build_dataset_and_sampler(
                target_dir,
                cache_dir,
                "train",
                train_csv,
                train_data,
                num_speaker,
                frame_shift,
                build_dataset,
                build_batch_sampler,
            )
            valid_ds, valid_bs = self._build_dataset_and_sampler(
                target_dir,
                cache_dir,
                "valid",
                valid_csv,
                valid_data,
                num_speaker,
                frame_shift,
                build_dataset,
                build_batch_sampler,
            )

            build_model_all_args = dict(
                build_model=build_model,
                model_output_size=model_output_size,
                build_upstream=build_upstream,
                build_featurizer=build_featurizer,
                build_downstream=build_downstream,
            )
            build_task_all_args_except_model = dict(
                build_task=build_task,
            )

            self.train(
                train,
                train_dir,
                build_model_all_args,
                build_task_all_args_except_model,
                save_model,
                save_task,
                build_optimizer,
                build_scheduler,
                evaluate,
                train_ds,
                train_bs,
                self.build_collate_fn(build_collate_fn, "train"),
                valid_ds,
                valid_bs,
                self.build_collate_fn(build_collate_fn, "valid"),
                device=device,
                eval_batch=eval_batch,
                num_workers=num_workers,
                world_size=world_size,
                rank=rank,
            )

        def check_fn():
            assert (train_dir / "valid_best").is_dir()

        self._stage_check(stage_id, stop, check_fn)

        stage_id = 2
        test_ckpt_dir: Path = Path(test_ckpt_dir or target_dir / "train" / "valid_best")
        test_dirs = []
        for test_idx, test_data in enumerate(test_datas):
            test_name = Path(test_data).stem
            test_dir: Path = (
                target_dir
                / "evaluate"
                / test_ckpt_dir.relative_to(train_dir).as_posix().replace("/", "-")
                / test_name
            )
            test_dirs.append(test_dir)

        if start <= stage_id:
            logger.info(f"Stage {stage_id}: Test model: {test_ckpt_dir}")
            for test_idx, test_data in enumerate(test_datas):
                test_csv = test_csvs[test_idx]
                test_dir = test_dirs[test_idx]
                test_dir.mkdir(exist_ok=True, parents=True)

                logger.info(
                    f"Stage {stage_id}.{test_idx}: Test model on {test_dir} and dump prediction"
                )
                test_ds, test_bs = self._build_dataset_and_sampler(
                    target_dir,
                    cache_dir,
                    "test",
                    test_csv,
                    test_data,
                    num_speaker,
                    frame_shift,
                    build_dataset,
                    build_batch_sampler,
                )

                _, valid_best_task = self.load_model_and_task(test_ckpt_dir)
                logs: dict = self.evaluate(
                    evaluate,
                    "test",
                    valid_best_task,
                    test_ds,
                    test_bs,
                    self.build_collate_fn(build_collate_fn, "test"),
                    eval_batch,
                    test_dir,
                    device,
                    num_workers,
                )
                test_metrics = {name: float(value) for name, value in logs.items()}
                with (test_dir / f"result.yaml").open("w") as f:
                    yaml.safe_dump(test_metrics, f)

        def check_fn():
            for test_dir in test_dirs:
                assert (test_dir / "prediction").is_dir()

        self._stage_check(stage_id, stop, check_fn)

        stage_id = 3
        if start <= stage_id:
            logger.info(f"Stage {stage_id}: Score model: {test_ckpt_dir}")
            self.scoring(scoring, stage_id, test_dirs, test_rttms, frame_shift)

        return stage_id

    def scoring(
        self,
        scoring: dict,
        stage_id: int,
        test_dirs: List[str],
        test_rttms: List[str],
        frame_shift: int,
    ):
        """
        Score the prediction

        Args:
            scoring (dict):

                ====================  ====================
                key                   description
                ====================  ====================
                thresholds            (List[int]) - Given the 0~1 (float) soft prediction, the threshold decides \
                                        how to get the 0/1 hard prediction. This list are all the thresholds to try.
                median_filters        (List[int]) - After getting hard prediction, use median filter to smooth out the \
                                        prediction. This list are all the median filter sizes to try.
                ====================  ====================

            *others:
                This method is not designed to be overridden
        """

        @dataclass
        class ScoreConfig:
            thresholds: List[int]
            median_filters: List[int]

        conf = ScoreConfig(**scoring)

        for test_idx, test_dir in enumerate(test_dirs):
            logger.info(
                f"Stage {stage_id}.{test_idx}: Make RTTM and Score from prediction"
            )
            best_der, (best_th, best_med) = make_rttm_and_score(
                test_dir / "prediction",
                test_dir / "score",
                test_rttms[test_idx],
                frame_shift,
                conf.thresholds,
                conf.median_filters,
            )

            logger.info(f"Best dscore DER: {best_der}")
            with (test_dir / "dscore.yaml").open("w") as f:
                yaml.safe_dump(
                    dict(
                        der=best_der,
                        threshold=best_th,
                        median_filter=best_med,
                    ),
                    f,
                )

    def _build_dataset_and_sampler(
        self,
        target_dir: str,
        cache_dir: str,
        mode: str,
        data_csv: str,
        data_dir: str,
        num_speakers: int,
        frame_shift: int,
        build_dataset: dict,
        build_batch_sampler: dict,
    ):
        logger.info(f"Build {mode} dataset")
        dataset = self.build_dataset(
            build_dataset,
            target_dir,
            cache_dir,
            mode,
            data_csv,
            data_dir,
            num_speakers,
            frame_shift,
        )
        logger.info(f"Build {mode} batch sampler")
        batch_sampler = self.build_batch_sampler(
            build_batch_sampler,
            target_dir,
            cache_dir,
            mode,
            data_csv,
            data_dir,
            dataset,
        )
        return dataset, batch_sampler

    def build_task(self, build_task: dict, model):
        """
        Build the task, which defines the logics for every train/valid/test forward step for the :code:`model`,
        and the logics for how to reduce all the batch results from multiple train/valid/test steps into metrics

        By default build :obj:`DiarizationPIT`

        Args:
            build_task (dict): same in :obj:`default_config`, no argument supported for now
            model (torch.nn.Module): the model built by :obj:`build_model`

        Returns:
            Task
        """
        task = DiarizationPIT(model)
        return task



import tempfile
from pathlib import Path
from typing import List

# fmt: off
# isort: off
import s3prl
import torch
from s3prl.upstream.multires_hubert.hubert_model import (
    MultiresHubertConfig,
    MultiresHubertModel,
    MultiresHubertPretrainingConfig,
)
from s3prl.upstream.utils import load_fairseq_ckpt, merge_with_parent
from s3prl.util.download import _urls_to_filepaths
# isort: on
# fmt: on


def load_and_convert_fairseq_ckpt(fairseq_source: str, output_path: str = None):
    from fairseq.data.dictionary import Dictionary

    state, cfg = load_fairseq_ckpt(fairseq_source)

    dicts: List[Dictionary] = state["task_state"]["dictionaries"]
    symbols = [dictionary.symbols for dictionary in dicts]

    output_state = {
        "task_cfg": cfg["task"],
        "model_cfg": cfg["model"],
        "model_weight": state["model"],
        "dictionaries_symbols": symbols,
    }

    if output_path is not None:
        Path(output_path).parent.mkdir(exist_ok=True, parents=True)
        torch.save(output_state, output_path)


def load_converted_model(ckpt: str):
    ckpt_state = torch.load(ckpt, map_location="cpu")

    for required_key in [
        "task_cfg",
        "model_cfg",
        "model_weight",
        "dictionaries_symbols",
    ]:
        if required_key not in ckpt_state:
            raise ValueError(
                f"{ckpt} is not a valid checkpoint since the required key: {required_key} is missing"
            )

    task_cfg = merge_with_parent(
        MultiresHubertPretrainingConfig, ckpt_state["task_cfg"]
    )
    model_cfg = merge_with_parent(MultiresHubertConfig, ckpt_state["model_cfg"])
    model = MultiresHubertModel(model_cfg, task_cfg, ckpt_state["dictionaries_symbols"])

    model.load_state_dict(ckpt_state["model_weight"])
    return model, task_cfg


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("fairseq_ckpt")
    parser.add_argument(
        "--output_dir", default=Path(s3prl.__file__).parent.parent / "converted_ckpts"
    )
    args = parser.parse_args()

    Path(args.output_dir).parent.mkdir(exist_ok=True, parents=True)
    load_and_convert_fairseq_ckpt(
        args.fairseq_ckpt, Path(args.output_dir) / f"{Path(args.fairseq_ckpt).stem}.pt"
    )



from .base import SequentialDataPipe
from .common_pipes import EncodeText, GenerateTokenizer, LoadAudio, SetOutputKeys


class Speech2TextPipe(SequentialDataPipe):
    """
    each item in the input dataset should have:
        wav_path: str
        transcription: str
    """

    def __init__(
        self,
        generate_tokenizer: bool = False,
        vocab_type: str = "character",
        text_file: str = None,
        vocab_file: str = None,
        slots_file: str = None,
        vocab_args: dict = None,
    ):
        output_keys = dict(
            x="wav",
            x_len="wav_len",
            labels="transcription",
            class_ids="tokenized_text",
            unique_name="id",
        )

        super().__init__(
            LoadAudio(),
            GenerateTokenizer(
                generate=generate_tokenizer,
                vocab_type=vocab_type,
                text_file=text_file,
                vocab_file=vocab_file,
                slots_file=slots_file,
                vocab_args=vocab_args,
            ),
            EncodeText(),
            SetOutputKeys(output_keys=output_keys),
        )



# --------------------------------------------------------
# LightHuBERT: Lightweight and Configurable Speech Representation Learning with Once-for-All Hidden-Unit BERT (https://arxiv.org/pdf/2203.15610.pdf)
# Github source: https://github.com/mechanicalsea/lighthubert
# Licensed under The MIT License [see LICENSE for details]
# Based on fairseq code bases
# https://github.com/pytorch/fairseq
# --------------------------------------------------------

import math
from typing import List

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F

from ..functional.fairseq_utils import get_activation_fn, index_put, pad_to_multiple
from .fairseq_modules import SamePad, TransposeLast
from .scaling_conv import SConv1d
from .scaling_layernorm import SLayerNorm
from .scaling_linear import SLinear
from .scaling_multihead import SMHA


class STransformerSentenceEncoderLayer(nn.Module):
    """TransformerSentenceEncoderLayer: variable input (i.e., output) size, attention heads, and ffn embedding.
    The `fc1` and `fc2` enable sharing weights with FFN ratio, e.g., 2.5, 3.0, 3.5, 4.0

    Dynamic: self.self_attn, self.fc1, self.fc2, self.self_attn_layer_norm, self.final_layer_norm

    wav2vec2:TransformerSentenceEncoderLayer(
        embedding_dim=self.embedding_dim,
        ffn_embedding_dim=args.encoder_ffn_embed_dim,
        num_attention_heads=args.encoder_attention_heads,
        dropout=self.dropout,
        attention_dropout=args.attention_dropout,
        activation_dropout=args.activation_dropout,
        activation_fn=args.activation_fn,
        layer_norm_first=args.layer_norm_first,
    )
    """

    def __init__(
        self,
        embedding_dim: float = 768,
        ffn_embedding_dim: float = 3072,
        num_attention_heads: float = 8,
        dropout: float = 0.1,
        attention_dropout: float = 0.1,
        activation_dropout: float = 0.1,
        activation_fn: str = "relu",
        layer_norm_first: bool = False,
    ) -> None:
        super().__init__()

        # Initialize parameters
        self.embedding_dim = embedding_dim
        self.dropout = dropout
        self.activation_dropout = activation_dropout

        # Initialize blocks
        self.activation_fn = get_activation_fn(activation_fn)
        self.self_attn = SMHA(
            self.embedding_dim,
            num_attention_heads,
            dropout=attention_dropout,
            self_attention=True,
        )  # lighthubert component

        self.dropout1 = nn.Dropout(dropout)
        self.dropout2 = nn.Dropout(self.activation_dropout)
        self.dropout3 = nn.Dropout(dropout)

        self.layer_norm_first = layer_norm_first

        # layer norm associated with the self attention layer
        self.self_attn_layer_norm = SLayerNorm(
            self.embedding_dim
        )  # lighthubert component
        self.fc1 = SLinear(
            self.embedding_dim,
            ffn_embedding_dim,
            in_splits=1,
            out_splits=ffn_embedding_dim // self.embedding_dim,
        )  # lighthubert component
        self.fc2 = SLinear(
            ffn_embedding_dim,
            self.embedding_dim,
            in_splits=ffn_embedding_dim // self.embedding_dim,
            out_splits=1,
        )  # lighthubert component

        # layer norm associated with the position wise feed-forward NN
        self.final_layer_norm = SLayerNorm(self.embedding_dim)  # lighthubert component

        # scaling module
        self.sample_atten_dim = None
        self.sample_embed_dim = None
        self.sample_ffn_embed = None
        self.sample_heads_num = None
        self.sample_attn_swz = "global"
        # replace self_attn, self_attn_layer_norm, fc1, fc2, final_layer_norm

    def forward(
        self,
        x: torch.Tensor,
        self_attn_mask: torch.Tensor = None,
        self_attn_padding_mask: torch.Tensor = None,
        need_weights: bool = False,
        att_args=None,
    ):
        """
        LayerNorm is applied either before or after the self-attention/ffn
        modules similar to the original Transformer imlementation.
        """
        residual = x

        if self.layer_norm_first:
            x = self.self_attn_layer_norm(x)
            x, attn = self.self_attn(
                query=x,
                key=x,
                value=x,
                key_padding_mask=self_attn_padding_mask,
                attn_mask=self_attn_mask,
                need_weights=False,
            )
            x = self.dropout1(x)
            x = residual + x

            residual = x
            x = self.final_layer_norm(x)
            x = self.activation_fn(self.fc1(x))
            x = self.dropout2(x)
            x = self.fc2(x)

            layer_result = x

            x = self.dropout3(x)
            x = residual + x
        else:
            x, attn = self.self_attn(
                query=x,
                key=x,
                value=x,
                key_padding_mask=self_attn_padding_mask,
                need_weights=False,
            )

            x = self.dropout1(x)
            x = residual + x

            x = self.self_attn_layer_norm(x)

            residual = x
            x = self.activation_fn(self.fc1(x))
            x = self.dropout2(x)
            x = self.fc2(x)

            layer_result = x

            x = self.dropout3(x)
            x = residual + x
            x = self.final_layer_norm(x)

        return x, (attn, layer_result)

    def set_sample_config(
        self,
        sample_atten_dim: int,
        sample_embed_dim: int,
        sample_ffn_embed: int,
        sample_heads_num: int,
        sample_sliding_attn_window="global",
    ):
        assert (
            sample_atten_dim == sample_heads_num * 64
        ), f"{sample_atten_dim}-{sample_heads_num}"
        self.sample_atten_dim = sample_atten_dim
        self.sample_embed_dim = sample_embed_dim
        self.sample_ffn_embed = sample_ffn_embed
        self.sample_heads_num = sample_heads_num
        self.sample_attn_swz = sample_sliding_attn_window
        self._sample_parameters()

    def _sample_parameters(self):
        self.self_attn.set_sample_config(
            self.sample_atten_dim,
            self.sample_heads_num,
            self.sample_embed_dim,
            self.sample_attn_swz,
        )
        self.self_attn_layer_norm.set_sample_config(self.sample_embed_dim)
        self.fc1.set_sample_config(
            self.sample_embed_dim,
            self.sample_ffn_embed,
            1,
            self.sample_ffn_embed / self.sample_embed_dim,
        )
        self.fc2.set_sample_config(
            self.sample_ffn_embed,
            self.sample_embed_dim,
            self.sample_ffn_embed / self.sample_embed_dim,
            1,
        )
        self.final_layer_norm.set_sample_config(self.sample_embed_dim)

    @property
    def sample_modules(self):
        return {
            "self_attn": SMHA,
            "self_attn_layer_norm": SLayerNorm,
            "fc1": SLinear,
            "fc2": SLinear,
            "final_layer_norm": SLayerNorm,
        }

    def calc_sampled_param_num(self):
        total_params = 0
        for module_name in self.sample_modules.keys():
            m = getattr(self, module_name)
            total_params += m.calc_sampled_param_num()
        return total_params

    def get_complexity(self, sequence_length):
        total_flops = 0
        for module_name in self.sample_modules.keys():
            m = getattr(self, module_name)
            total_flops += m.get_complexity(sequence_length)
        # 2 residual connection
        total_flops += 2 * sequence_length * self.sample_embed_dim
        # activation
        total_flops += sequence_length * self.sample_ffn_embed
        return total_flops


def make_sconv_pos(e, k, g):
    pos_conv = SConv1d(
        e,
        e,
        kernel_size=k,
        padding=k // 2,
        groups=g,
    )
    dropout = 0
    std = math.sqrt((4 * (1.0 - dropout)) / (k * e))
    nn.init.normal_(pos_conv.weight, mean=0, std=std)
    nn.init.constant_(pos_conv.bias, 0)

    pos_conv = nn.utils.weight_norm(pos_conv, name="weight", dim=2)
    pos_conv = nn.Sequential(pos_conv, SamePad(k), nn.GELU())

    return pos_conv


def init_sbert_params(module):
    """
    Initialize the weights specific to the BERT Model.
    This overrides the default initializations depending on the specified arguments.
        1. If normal_init_linear_weights is set then weights of scaling_linear
           layer will be initialized using the normal distribution and
           bais will be set to the specified value.
        1. If normal_init_proj_weights is set then weights of
           in_project_weight for scaling_MultiHeadAttention initialized using
           the normal distribution (to be validated).
    """

    def normal_(data):
        # with FSDP, module params will be on CUDA, so we cast them back to CPU
        # so that the RNG is consistent with and without FSDP
        data.copy_(data.cpu().normal_(mean=0.0, std=0.02).to(data.device))

    if isinstance(module, SLinear):
        normal_(module.weight.data)
        if module.bias is not None:
            module.bias.data.zero_()
    if isinstance(module, SMHA):
        normal_(module.q_proj.weight.data)
        normal_(module.k_proj.weight.data)
        normal_(module.v_proj.weight.data)


class STransformerEncoder(nn.Module):
    """TransformerEncoder: variable layers number, input dim (or outout),
    heads numbers, and ffn embedding dim.

    Note that `input dim (or outout)` now is fixed to be subject
    to hidden representations of teacher model.

    Dynamic: self.layers, self.pos_conv.0, self.layer_norm

    wav2vec2:TransformerEncoder(args)
    """

    def build_sencoder_layer(self, args):
        assert args.layer_type == "transformer", args.layer_type
        if args.layer_type == "transformer":
            layer = STransformerSentenceEncoderLayer(
                embedding_dim=self.embedding_dim,
                ffn_embedding_dim=args.encoder_ffn_embed_dim,
                num_attention_heads=args.encoder_attention_heads,
                dropout=self.dropout,
                attention_dropout=args.attention_dropout,
                activation_dropout=args.activation_dropout,
                activation_fn=args.activation_fn,
                layer_norm_first=args.layer_norm_first,
            )
        return layer

    def __init__(self, args):
        super().__init__()

        self.dropout = args.dropout
        self.embedding_dim = args.encoder_embed_dim
        self.required_seq_len_multiple = args.required_seq_len_multiple

        pos_conv_depth = getattr(args, "pos_conv_depth", 1)
        if pos_conv_depth > 1:
            num_layers = args.pos_conv_depth
            k = max(3, args.conv_pos // num_layers)

            def make_sconv_block(e, k, g, l):
                return nn.Sequential(
                    *[
                        nn.Sequential(
                            SConv1d(
                                e,
                                e,
                                kernel_size=k,
                                padding=k // 2,
                                groups=g,
                            ),
                            SamePad(k),
                            TransposeLast(),
                            SLayerNorm(e, elementwise_affine=False),
                            TransposeLast(),
                            nn.GELU(),
                        )
                        for _ in range(l)
                    ]
                )

            self.pos_conv = make_sconv_block(
                self.embedding_dim, k, args.conv_pos_groups, num_layers
            )

        else:
            self.pos_conv = make_sconv_pos(
                self.embedding_dim,
                args.conv_pos,
                args.conv_pos_groups,
            )

        self.layers = nn.ModuleList(
            [self.build_sencoder_layer(args) for _ in range(args.encoder_layers)]
        )
        self.layer_norm_first = args.layer_norm_first
        self.layer_norm = SLayerNorm(self.embedding_dim)
        self.layerdrop = args.encoder_layerdrop

        self.apply(init_sbert_params)

        self.args = args
        assert args.encoder_layers in [12], f"encoder layers: {args.encoder_layers}"
        # layer block: 1-4, 5-8, 9-12
        self.depth_maps = {
            6: [1, 1, 0, 0, 1, 1, 0, 0, 1, 1, 0, 0],
            7: [1, 1, 1, 0, 1, 1, 0, 0, 1, 1, 0, 0],
            8: [1, 1, 1, 0, 1, 1, 1, 0, 1, 1, 0, 0],
            9: [1, 1, 1, 0, 1, 1, 1, 0, 1, 1, 1, 0],
            10: [1, 1, 1, 1, 1, 1, 1, 0, 1, 1, 1, 0],
            11: [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0],
            12: [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1],
        }  # 1: keep layer, 0: drop layer
        self.sample_layer_num = args.encoder_layers
        self.sample_atten_dim = None
        self.sample_embed_dim = None
        self.sample_ffn_embed = None
        self.sample_heads_num = None
        self.sample_attn_swz = None

    def forward(self, x, padding_mask=None, layer=None):
        x, layer_results = self.extract_features(x, padding_mask, layer)

        if self.layer_norm_first and layer is None:
            x = self.layer_norm(x)

        return x, layer_results

    def extract_features(
        self,
        x,
        padding_mask=None,
        tgt_layer=None,
        min_layer=0,
    ):
        """Refactor it to enable varible layers number"""
        if padding_mask is not None:
            x = index_put(x, padding_mask, 0)

        x_conv = self.pos_conv(x.transpose(1, 2))
        x_conv = x_conv.transpose(1, 2)
        x = x + x_conv

        if not self.layer_norm_first:
            x = self.layer_norm(x)

        # pad to the sequence length dimension
        x, pad_length = pad_to_multiple(
            x, self.required_seq_len_multiple, dim=-2, value=0
        )
        if pad_length > 0 and padding_mask is None:
            padding_mask = x.new_zeros((x.size(0), x.size(1)), dtype=torch.bool)
            padding_mask[:, -pad_length:] = True
        else:
            padding_mask, _ = pad_to_multiple(
                padding_mask, self.required_seq_len_multiple, dim=-1, value=True
            )
        x = F.dropout(x, p=self.dropout, training=self.training)

        # B x T x C -> T x B x C
        x = x.transpose(0, 1)

        layer_results = []
        layerdrop_index = self.depth_maps[self.sample_layer_num]
        r = None
        for i, layer in enumerate(self.layers):
            dropout_probability = np.random.random()
            if layerdrop_index[i] == 1 and (
                not self.training or dropout_probability > self.layerdrop
            ):
                x, (z, lr) = layer(
                    x, self_attn_padding_mask=padding_mask, need_weights=False
                )
                if i >= min_layer:
                    layer_results.append((x, z, lr))
            if i == tgt_layer:
                r = x
                break

        if r is not None:
            x = r

        # T x B x C -> B x T x C
        x = x.transpose(0, 1)

        # undo paddding
        if pad_length > 0:
            x = x[:, :-pad_length]

            def undo_pad(a, b, c):
                return (
                    a[:-pad_length],
                    b[:-pad_length] if b is not None else b,
                    c[:-pad_length],
                )

            layer_results = [undo_pad(*u) for u in layer_results]

        return x, layer_results

    def max_positions(self):
        """Maximum output length supported by the encoder."""
        return self.args.max_positions

    def set_sample_config(
        self,
        sample_layer_num: int,
        sample_atten_dim: List[int],
        sample_embed_dim: int,
        sample_ffn_embed: List[int],
        sample_heads_num: List[int],
        sample_attn_swz: List,
    ):
        self.sample_layer_num = sample_layer_num
        self.sample_atten_dim = sample_atten_dim
        self.sample_embed_dim = sample_embed_dim
        self.sample_ffn_embed = sample_ffn_embed
        self.sample_heads_num = sample_heads_num
        self.sample_attn_swz = sample_attn_swz
        self._sample_parameters()

    def _sample_parameters(self):
        if getattr(self.args, "pos_conv_depth", 1) > 1:
            pos_conv_depth = getattr(self.args, "pos_conv_depth", 1)
            prune_encoder_pos_conv = getattr(self.args, "prune_encoder_pos_conv", True)
            sample_input_size = (
                self.sample_embed_dim
                if prune_encoder_pos_conv
                else self.args.encoder_embed_dim
            )
            for i, mi in enumerate(self.pos_conv):
                if i == 0:
                    mi[0].set_sample_config(self.sample_embed_dim, sample_input_size)
                    mi[3].set_sample_config(sample_input_size)
                elif i < pos_conv_depth - 1:
                    mi[0].set_sample_config(sample_input_size, sample_input_size)
                    mi[3].set_sample_config(sample_input_size)
                else:
                    mi[0].set_sample_config(sample_input_size, self.sample_embed_dim)
                    mi[3].set_sample_config(self.sample_embed_dim)
        else:
            self.pos_conv[0].set_sample_config(
                self.sample_embed_dim,
                self.sample_embed_dim,
            )
        layerdrop_index = self.depth_maps[self.sample_layer_num]
        for i, layer in enumerate(self.layers):
            if layerdrop_index[i] == 1:
                # enable layer-drop training
                layer_i = sum(layerdrop_index[: i + 1]) - 1
                layer.set_sample_config(
                    self.sample_atten_dim[layer_i],
                    self.sample_embed_dim,
                    self.sample_ffn_embed[layer_i],
                    self.sample_heads_num[layer_i],
                    self.sample_attn_swz[layer_i],
                )
        self.layer_norm.set_sample_config(self.sample_embed_dim)

    def calc_sampled_param_num(self):
        total_params = 0
        layerdrop_index = self.depth_maps[self.sample_layer_num]
        total_params += sum(
            layer.calc_sampled_param_num()
            for i, layer in enumerate(self.layers)
            if layerdrop_index[i] == 1
        )
        if getattr(self.args, "pos_conv_depth", 1) > 1:
            for mi in self.pos_conv:
                total_params += mi[0].calc_sampled_param_num()
        else:
            total_params += self.pos_conv[0].calc_sampled_param_num()
        total_params += self.layer_norm.calc_sampled_param_num()
        return total_params

    def get_complexity(self, sequence_length):
        total_flops = 0
        total_flops += sum(
            layer.get_complexity(sequence_length) for layer in self.layers
        )
        # pos_conv
        if getattr(self.args, "pos_conv_depth", 1) > 1:
            for mi in self.pos_conv:
                total_flops += (
                    self.sample_embed_dim
                    * sequence_length
                    * (
                        self.sample_embed_dim / mi[0].groups * mi[0].kernel_size
                        + int(mi[0].bias is not None)
                    )
                )  # conv
                total_flops += self.sample_embed_dim * sequence_length  # layernorm
        else:
            total_flops += (
                self.sample_embed_dim
                * sequence_length
                * (
                    self.sample_embed_dim
                    / self.pos_conv[0].groups
                    * self.pos_conv[0].kernel_size
                    + int(self.pos_conv[0].bias is not None)
                )
            )
        # layer_norm
        total_flops += self.sample_embed_dim * sequence_length
        return total_flops



"""
Metrics for diarization

Authors:
  * Jiatong Shi 2021
"""

from itertools import permutations

import numpy as np
import torch

__all__ = [
    "calc_diarization_error",
]


def calc_diarization_error(pred, label, length):
    (batch_size, max_len, num_output) = label.size()
    # mask the padding part
    mask = np.zeros((batch_size, max_len, num_output))
    for i in range(batch_size):
        mask[i, : length[i], :] = 1

    # pred and label have the shape (batch_size, max_len, num_output)
    label_np = label.data.cpu().numpy().astype(int)
    pred_np = (pred.data.cpu().numpy() > 0).astype(int)
    label_np = label_np * mask
    pred_np = pred_np * mask
    length = length.data.cpu().numpy()

    # compute speech activity detection error
    n_ref = np.sum(label_np, axis=2)
    n_sys = np.sum(pred_np, axis=2)
    speech_scored = float(np.sum(n_ref > 0))
    speech_miss = float(np.sum(np.logical_and(n_ref > 0, n_sys == 0)))
    speech_falarm = float(np.sum(np.logical_and(n_ref == 0, n_sys > 0)))

    # compute speaker diarization error
    speaker_scored = float(np.sum(n_ref))
    speaker_miss = float(np.sum(np.maximum(n_ref - n_sys, 0)))
    speaker_falarm = float(np.sum(np.maximum(n_sys - n_ref, 0)))
    n_map = np.sum(np.logical_and(label_np == 1, pred_np == 1), axis=2)
    speaker_error = float(np.sum(np.minimum(n_ref, n_sys) - n_map))
    correct = float(1.0 * np.sum((label_np == pred_np) * mask) / num_output)
    num_frames = np.sum(length)
    return (
        correct,
        num_frames,
        speech_scored,
        speech_miss,
        speech_falarm,
        speaker_scored,
        speaker_miss,
        speaker_falarm,
        speaker_error,
    )



"""
The model architecture of VQ-APC

Authors:
  * Andy T. Liu 2022
"""


import numpy as np
import torch
import torch.nn as nn
from torch.nn.functional import gumbel_softmax

from s3prl import Output

EPS = 1e-10

__all__ = [
    "VqApcLayer",
]


class VqApcLayer(nn.Module):
    """
    The Vq Layer.
    Currently used in the upstream model of VQ-APC (nn/rnn_apc.py).
    Defines a VQ layer that follows an RNN layer.
    """

    def __init__(self, input_size, codebook_size, code_dim, gumbel_temperature):
        """
        Args:
            input_size (int):
                An int indicating the pre-quantized input feature size,
                usually the hidden size of RNN.
            codebook_size (int):
                An int indicating the number of codes.
            code_dim (int):
                An int indicating the size of each code. If not the last layer,
                then must equal to the RNN hidden size.
            gumbel_temperature (float):
                A float indicating the temperature for gumbel-softmax.
        """
        super(VqApcLayer, self).__init__()
        # Directly map to logits without any transformation.
        self.codebook_size = codebook_size
        self.vq_logits = nn.Linear(input_size, codebook_size)
        self.gumbel_temperature = gumbel_temperature
        self.codebook_CxE = nn.Linear(codebook_size, code_dim, bias=False)
        self.token_usg = np.zeros(codebook_size)

    def forward(self, inputs_BxLxI, testing=False):
        """
        Args:
            inputs_BxLxI (torch.LongTensor):
                A 3d-tensor representing the input features.
            testing (bool):
                A bool indicating training or testing phase.
                Default: False
        Return:
            Output (s3prl.Output):
                An Output module that contains `output` and `logit`

                output (codes_BxLxE):
                    The VQ codes.
                logit (logits_BxLxC):
                    The VQ logits.
        """
        logits_BxLxC = self.vq_logits(inputs_BxLxI)
        if testing:
            # During inference, just take the max index.
            shape = logits_BxLxC.size()
            _, ind = logits_BxLxC.max(dim=-1)
            onehot_BxLxC = torch.zeros_like(logits_BxLxC).view(-1, shape[-1])
            onehot_BxLxC.scatter_(1, ind.view(-1, 1), 1)
            onehot_BxLxC = onehot_BxLxC.view(*shape)
        else:
            onehot_BxLxC = gumbel_softmax(
                logits_BxLxC, tau=self.gumbel_temperature, hard=True, eps=EPS, dim=-1
            )
            self.token_usg += (
                onehot_BxLxC.detach()
                .cpu()
                .reshape(-1, self.codebook_size)
                .sum(dim=0)
                .numpy()
            )
        codes_BxLxE = self.codebook_CxE(onehot_BxLxC)

        return Output(output=codes_BxLxE, logit=logits_BxLxC)

    def report_ppx(self):
        """
        Computes perplexity of distribution over codebook.
        """
        acc_usg = self.token_usg / sum(self.token_usg)
        return 2 ** sum(-acc_usg * np.log2(acc_usg + EPS))

    def report_usg(self):
        """
        Computes usage each entry in codebook.
        """
        acc_usg = self.token_usg / sum(self.token_usg)
        # Reset
        self.token_usg = np.zeros(self.codebook_size)
        return acc_usg



from s3prl.util.download import _urls_to_filepaths

from .expert import UpstreamExpert as _UpstreamExpert


def byol_s_default(refresh: bool = False, **kwds):
    kwds["model_name"] = "default"
    kwds["ckpt"] = _urls_to_filepaths(
        "https://github.com/GasserElbanna/serab-byols/raw/main/checkpoints/default2048_BYOLAs64x96-2105311814-e100-bs256-lr0003-rs42.pth",
        refresh=refresh,
    )
    return _UpstreamExpert(**kwds)


def byol_s_cvt(refresh: bool = False, **kwds):
    kwds["model_name"] = "cvt"
    kwds["ckpt"] = _urls_to_filepaths(
        "https://github.com/GasserElbanna/serab-byols/raw/main/checkpoints/cvt_s1-d1-e64_s2-d1-e256_s3-d1-e512_BYOLAs64x96-osandbyolaloss6373-e100-bs256-lr0003-rs42.pth",
        refresh=refresh,
    )
    return _UpstreamExpert(**kwds)


def byol_s_resnetish34(refresh: bool = False, **kwds):
    kwds["model_name"] = "resnetish34"
    kwds["ckpt"] = _urls_to_filepaths(
        "https://github.com/GasserElbanna/serab-byols/raw/main/checkpoints/resnetish34_BYOLAs64x96-2105271915-e100-bs256-lr0003-rs42.pth",
        refresh=refresh,
    )
    return _UpstreamExpert(**kwds)



from s3prl.util.download import _urls_to_filepaths

from .expert import UpstreamExpert as _UpstreamExpert


def espnet_hubert_custom(ckpt, *args, config=None, **kwargs):
    return _UpstreamExpert(ckpt, *args, **kwargs)


def espnet_hubert_local(*args, **kwargs):
    return espnet_hubert_custom(*args, **kwargs)


def cvhubert(*args, refresh=False, **kwargs):
    url = "https://huggingface.co/espnet/espnet_cvhubert/resolve/main/exp/hubert_iter2_train_ssl_torchaudiohubert_base_960h_pretrain_it2_raw/latest.pth"
    config_url = "https://huggingface.co/espnet/espnet_cvhubert/raw/main/exp/hubert_iter2_train_ssl_torchaudiohubert_base_960h_pretrain_it2_raw/config.yaml"
    ckpt, config = _urls_to_filepaths(url, config_url, refresh=refresh)
    return espnet_hubert_custom(ckpt, config)


def wavlablm_ek_40k(*args, refresh=False, **kwargs):
    url = "https://huggingface.co/espnet/WavLabLM-EK-40k/resolve/main/exp_li/hubert_iter2_train_ssl_torchaudiohubert_large_960h_pretrain_it2_cont_raw_layer_9/5epoch.pth"
    config_url = "https://huggingface.co/espnet/WavLabLM-EK-40k/raw/main/exp_li/hubert_iter2_train_ssl_torchaudiohubert_large_960h_pretrain_it2_cont_raw_layer_9/config.yaml"
    ckpt, config = _urls_to_filepaths(url, config_url, refresh=refresh)
    return espnet_hubert_custom(ckpt, config)


def wavlablm_ms_40k(*args, refresh=False, **kwargs):
    url = "https://huggingface.co/espnet/WavLabLM-MS-40k/resolve/main/exp_babel/hubert_iter2_train_ssl_torchaudiohubert_large_960h_pretrain_it2_wavlm_babel_light_raw_layer_9/5epoch.pth"
    config_url = "https://huggingface.co/espnet/WavLabLM-MS-40k/raw/main/exp_babel/hubert_iter2_train_ssl_torchaudiohubert_large_960h_pretrain_it2_wavlm_babel_light_raw_layer_9/config.yaml"
    ckpt, config = _urls_to_filepaths(url, config_url, refresh=refresh)
    return espnet_hubert_custom(ckpt, config)


def wavlablm_mk_40k(*args, refresh=False, **kwargs):
    url = "https://huggingface.co/espnet/WavLabLM-MK-40k/resolve/main/exp_li/hubert_iter2_train_ssl_torchaudiohubert_large_960h_pretrain_it2_wavlm_raw_layer_9/valid.acc_m.ave_10best.pth"
    config_url = "https://huggingface.co/espnet/WavLabLM-MK-40k/raw/main/exp_li/hubert_iter2_train_ssl_torchaudiohubert_large_960h_pretrain_it2_wavlm_raw_layer_9/config.yaml"
    ckpt, config = _urls_to_filepaths(url, config_url, refresh=refresh)
    return espnet_hubert_custom(ckpt, config)


def espnet_hubert_base_iter1(*args, refresh=False, **kwargs):
    url = "https://huggingface.co/espnet/simpleoier_librispeech_hubert_iter1_train_ssl_torchaudiohubert_base_960h_pretrain_it1_raw/resolve/main/exp/hubert_iter1_train_ssl_torchaudiohubert_base_960h_pretrain_it1_raw/valid.loss.ave.pth"
    config_url = "https://huggingface.co/espnet/simpleoier_librispeech_hubert_iter1_train_ssl_torchaudiohubert_base_960h_pretrain_it1_raw/raw/main/exp/hubert_iter1_train_ssl_torchaudiohubert_base_960h_pretrain_it1_raw/config.yaml"
    ckpt, config = _urls_to_filepaths(url, config_url, refresh=refresh)
    return espnet_hubert_custom(ckpt, config)


def espnet_hubert_base_iter0(*args, refresh=False, **kwargs):
    url = "https://huggingface.co/espnet/simpleoier_librispeech_hubert_iter0_train_ssl_torchaudiohubert_base_960h_pretrain_it0_raw/resolve/main/exp/hubert_iter0_train_ssl_torchaudiohubert_base_960h_pretrain_it0_raw/valid.loss.ave.pth"
    config_url = "https://huggingface.co/espnet/simpleoier_librispeech_hubert_iter0_train_ssl_torchaudiohubert_base_960h_pretrain_it0_raw/raw/main/exp/hubert_iter0_train_ssl_torchaudiohubert_base_960h_pretrain_it0_raw/config.yaml"
    ckpt, config = _urls_to_filepaths(url, config_url, refresh=refresh)
    return espnet_hubert_custom(ckpt, config)


def espnet_hubert_large_gs_ll60k(*args, refresh=False, **kwargs):
    url = "https://huggingface.co/espnet/hubert_large_gs_16_librilight60k/resolve/main/mnt/datastore/exp/hubert_iter1_train_ssl_torchaudiohubert_large_960h_pretrain_it2_bins_raw/valid.loss.ave_10best.pth"
    config_url = "https://huggingface.co/espnet/hubert_large_gs_16_librilight60k/blob/main/mnt/datastore/exp/hubert_iter1_train_ssl_torchaudiohubert_large_960h_pretrain_it2_bins_raw/config.yaml"
    ckpt, config = _urls_to_filepaths(url, config_url, refresh=refresh)
    return espnet_hubert_custom(ckpt, config)



import torch
import torch.nn.functional as F
from torch import nn


class PasstBasicWrapper(nn.Module):
    def __init__(
        self,
        mel: nn.Module,
        net: nn.Module,
        max_model_window=10000,
        timestamp_window=160,
        timestamp_hop=50,
        scene_hop=2500,
        scene_embedding_size=1295,
        timestamp_embedding_size=1295,
        mode="all",
    ):
        """
        @param mel: spectrogram extractor
        @param net: network module
        @param max_model_window: maximum clip length allowed by the model (milliseconds).
        @param timestamp_hop: the hop lengh for timestamp embeddings (milliseconds).
        @param scene_hop: the hop lengh for scene embeddings (milliseconds).
        @param scene_embedding_size:
        @param timestamp_embedding_size:
        @param mode: "all", "embed_only", "logits"
        """
        torch.nn.Module.__init__(self)
        self.mel = mel
        self.net = net
        self.device_proxy = nn.Parameter(torch.zeros((1)))
        self.sample_rate = mel.sr
        self.timestamp_window = int(timestamp_window * self.sample_rate / 1000)
        self.max_model_window = int(max_model_window * self.sample_rate / 1000)
        self.timestamp_hop = int(timestamp_hop * self.sample_rate / 1000)
        self.scene_hop = int(scene_hop * self.sample_rate / 1000)
        self.scene_embedding_size = scene_embedding_size
        self.timestamp_embedding_size = timestamp_embedding_size
        self.mode = mode

    def device(self):
        return self.device_proxy.device

    def forward(self, x):
        specs = self.mel(x)
        specs = specs.unsqueeze(1)
        x, features = self.net(specs)
        if self.mode == "all":
            embed = torch.cat([x, features], dim=1)
        elif self.mode == "embed_only":
            embed = features
        elif self.mode == "logits":
            embed = x
        else:
            raise RuntimeError(
                f"mode='{self.mode}' is not recognized not in: all, embed_only, logits"
            )
        return embed

    def get_scene_embeddings(self, audio):
        """
        audio: n_sounds x n_samples of mono audio in the range [-1, 1]. All sounds in a batch will be padded/trimmed to the same length.
        model: Loaded Model.
        Returns:
        embedding: A float32 Tensor with shape (n_sounds, model.scene_embedding_size).
        """
        n_sounds, n_samples = audio.shape
        if n_samples <= self.max_model_window:
            embed = self(audio.contiguous())
            return embed
        embeddings, timestamps = self.get_timestamp_embeddings(
            audio, window_size=self.max_model_window, hop=self.scene_hop
        )
        return embeddings.mean(axis=1)

    def get_timestamp_embeddings(
        self, audio: torch.Tensor, window_size=None, hop=None, pad=None
    ):
        """
        audio: n_sounds x n_samples of mono audio in the range [-1, 1]. All sounds in a batch will be padded/trimmed to the same length.
        model: Loaded Model.
        Returns:
        embedding: A float32 Tensor with shape (n_sounds, n_timestamps, model.timestamp_embedding_size).
        timestamps: A float32 Tensor with shape (`n_sounds, n_timestamps). Centered timestamps in milliseconds corresponding to each embedding in the output.
        """
        if hop is None:
            hop = self.timestamp_hop
        if window_size is None:
            window_size = self.timestamp_window
        if pad is None:
            pad = window_size // 2

        n_sounds, n_samples = audio.shape
        audio = audio.unsqueeze(1)  # n_sounds,1, (n_samples+pad*2)
        # print(audio.shape)
        padded = F.pad(audio, (pad, pad), mode="reflect")
        # print(padded.shape)
        padded = padded.unsqueeze(1)  # n_sounds,1, (n_samples+pad*2)
        # print(padded.shape)
        segments = (
            F.unfold(padded, kernel_size=(1, window_size), stride=(1, hop))
            .transpose(-1, -2)
            .transpose(0, 1)
        )
        timestamps = []
        embeddings = []
        for i, segment in enumerate(segments):
            timestamps.append(i)
            emb = self(segment)
            embeddings.append(emb)
        timestamps = torch.as_tensor(timestamps) * hop * 1000.0 / self.sample_rate

        embeddings = torch.stack(embeddings).transpose(
            0, 1
        )  # now n_sounds, n_timestamps, timestamp_embedding_size
        timestamps = timestamps.unsqueeze(0).expand(n_sounds, -1)

        return embeddings, timestamps

    def get_timestamp_mels(
        self, audio: torch.Tensor, window_size=None, hop=None, pad=None
    ):
        """
        audio: n_sounds x n_samples of mono audio in the range [-1, 1]. All sounds in a batch will be padded/trimmed to the same length.
        model: Loaded Model.
        Returns:
        embedding: A float32 Tensor with shape (n_sounds, n_timestamps, model.timestamp_embedding_size).
        timestamps: A float32 Tensor with shape (`n_sounds, n_timestamps). Centered timestamps in milliseconds corresponding to each embedding in the output.
        """
        if hop is None:
            hop = self.timestamp_hop
        if window_size is None:
            window_size = self.timestamp_window
        if pad is None:
            pad = window_size // 2

        n_sounds, n_samples = audio.shape
        audio = audio.unsqueeze(1)  # n_sounds,1, (n_samples+pad*2)
        # print(audio.shape)
        padded = F.pad(audio, (pad, pad), mode="reflect")
        # print(padded.shape)
        padded = padded.unsqueeze(1)  # n_sounds,1, (n_samples+pad*2)
        # print(padded.shape)
        segments = (
            F.unfold(padded, kernel_size=(1, window_size), stride=(1, hop))
            .transpose(-1, -2)
            .transpose(0, 1)
        )
        timestamps = []
        embeddings = []
        for i, segment in enumerate(segments):
            timestamps.append(i)
            embeddings.append(self.mel(segment).reshape(n_sounds, 128 * 6))
        timestamps = torch.as_tensor(timestamps) * hop * 1000.0 / self.sample_rate

        embeddings = torch.stack(embeddings).transpose(
            0, 1
        )  # now n_sounds, n_timestamps, timestamp_embedding_size
        timestamps = timestamps.unsqueeze(0).expand(n_sounds, -1)

        return embeddings, timestamps



import json
from collections import defaultdict
from pathlib import Path

import pandas as pd
from omegaconf import MISSING

from ._hear_util import resample_hear_corpus
from .hear_fsd import HearFSD

ESC50_NUM_FOLDS = 5

__all__ = ["HearESC50"]


def hear_scene_kfolds(
    target_dir: str,
    cache_dir: str,
    dataset_root: str,
    test_fold: int,
    num_folds: int,
    get_path_only: bool = False,
):
    assert test_fold < num_folds, (
        "test_fold id must be smaller than num_folds. "
        f"get test_fold={test_fold} and num_folds={num_folds}"
    )

    target_dir = Path(target_dir)
    train_csv = target_dir / "train.csv"
    valid_csv = target_dir / "valid.csv"
    test_csv = target_dir / "test.csv"

    if get_path_only:
        return train_csv, valid_csv, [test_csv]

    resample_hear_corpus(dataset_root, target_sr=16000)

    dataset_root = Path(dataset_root)
    wav_root: Path = dataset_root / "16000"

    def load_json(filepath):
        with open(filepath, "r") as fp:
            return json.load(fp)

    fold_metas = []
    fold_datas = []
    for fold_id in range(num_folds):
        meta = load_json(dataset_root / f"fold{fold_id:2d}.json".replace(" ", "0"))
        fold_metas.append(meta)

        data = defaultdict(list)
        for k in list(meta.keys()):
            wav_path = wav_root / f"fold{fold_id:2d}".replace(" ", "0") / k
            labels = meta[k]
            data["id"].append(k)
            data["wav_path"].append(wav_path)
            data["labels"].append(",".join([str(label).strip() for label in labels]))

        df = pd.DataFrame(data=data)
        fold_datas.append(df)

    test_id = test_fold
    valid_id = (test_fold + 1) % num_folds
    train_ids = [idx for idx in range(num_folds) if idx not in [test_id, valid_id]]

    test_data = fold_datas[test_id]
    valid_data = fold_datas[valid_id]
    train_data = []
    for idx in train_ids:
        train_data.append(fold_datas[idx])
    train_data = pd.concat(train_data)

    train_data.to_csv(train_csv, index=False)
    valid_data.to_csv(valid_csv, index=False)
    test_data.to_csv(test_csv, index=False)

    return train_csv, valid_csv, [test_csv]


class HearESC50(HearFSD):
    def default_config(self) -> dict:
        return dict(
            start=0,
            stop=None,
            target_dir=MISSING,
            cache_dir=None,
            remove_all_cache=False,
            prepare_data=dict(
                dataset_root=MISSING,
                test_fold=MISSING,
                num_folds=ESC50_NUM_FOLDS,
            ),
            build_batch_sampler=dict(
                train=dict(
                    batch_size=32,
                    shuffle=True,
                ),
                valid=dict(
                    batch_size=1,
                ),
                test=dict(
                    batch_size=1,
                ),
            ),
            build_upstream=dict(
                name=MISSING,
            ),
            build_featurizer=dict(
                layer_selections=None,
                normalize=False,
            ),
            build_downstream=dict(
                hidden_layers=2,
                pooling_type="MeanPooling",
            ),
            build_model=dict(
                upstream_trainable=False,
            ),
            build_task=dict(
                prediction_type="multiclass",
                scores=["top1_acc", "d_prime", "aucroc", "mAP"],
            ),
            build_optimizer=dict(
                name="Adam",
                conf=dict(
                    lr=1.0e-3,
                ),
            ),
            build_scheduler=dict(
                name="ExponentialLR",
                gamma=0.9,
            ),
            save_model=dict(),
            save_task=dict(),
            train=dict(
                total_steps=4000,
                log_step=100,
                eval_step=500,
                save_step=100,
                gradient_clipping=1.0,
                gradient_accumulate=4,
                valid_metric="top1_acc",
                valid_higher_better=True,
                auto_resume=True,
                resume_ckpt_dir=None,
            ),
            evaluate=dict(),
        )

    def prepare_data(
        self,
        prepare_data: dict,
        target_dir: str,
        cache_dir: str,
        get_path_only: bool = False,
    ):
        return hear_scene_kfolds(
            **self._get_current_arguments(flatten_dict="prepare_data")
        )
